; DOS-less diskette boot loader
; Embedded PC's ISA Bus: Chapter 1 -- Ed Nisley
; Assemble with Borland TASM

; Loads first data file in directory to address 1000:0100
;  - file must be contiguous!
;  - overall length must be less than 64K - 256 bytes
;  - ignores volume label and similar debris
; Blinks DTR output on a COM port if there's any error

; Includes the 55 AA marker in the last two bytes for compatibility
;  with Award BIOSes, which seem to require that value before booting.
; Program load point is 1000:0100 to avoid diskette DMA problems
;  ... the DMA buffers cannot cross a 64K byte physical memory boundary

              IDEAL
              MODEL     TINY

;--------------
; Set up a few constants

BLINK_ON      EQU       2                   ; units of 64K microseconds
BLINK_OFF     EQU       3

COMPORT       EQU       1                   ; which COM port to blink

;--------------
; Sort out the diskette parameters

              IF        DISKSIZE EQ 720
              DISPLAY   "Creating 720K diskette boot sector"
N_TRACKS      EQU       80
N_TRACKSIZE   EQU       9
N_HEADS       EQU       2
N_CLUSTERSIZE EQU       2
N_ROOTFILES   EQU       112
N_MEDIAID     EQU       0F9h
N_FATSIZE     EQU       3

              ELSEIF    DISKSIZE EQ 360
              DISPLAY   "Creating 360K diskette boot sector"
N_TRACKS      EQU       40
N_TRACKSIZE   EQU       9
N_HEADS       EQU       2
N_CLUSTERSIZE EQU       2
N_ROOTFILES   EQU       112
N_MEDIAID     EQU       0FDh
N_FATSIZE     EQU       2

              ELSEIF    DISKSIZE EQ 1200
              DISPLAY   "Creating 1.2M diskette boot sector"
N_TRACKS      EQU       80
N_TRACKSIZE   EQU       15
N_HEADS       EQU       2
N_CLUSTERSIZE EQU       1
N_ROOTFILES   EQU       224
N_MEDIAID     EQU       0F9h
N_FATSIZE     EQU       7

              ELSEIF    DISKSIZE EQ 1440
              DISPLAY   "Creating 1.44M diskette boot sector"
N_TRACKS      EQU       80
N_TRACKSIZE   EQU       18
N_HEADS       EQU       2
N_CLUSTERSIZE EQU       1
N_ROOTFILES   EQU       224
N_MEDIAID     EQU       0F0h
N_FATSIZE     EQU       9

              ELSE
              DISPLAY   "Invalid diskette size, check -ddisksize= option"

              ENDIF

N_SECTORS     =         N_TRACKS*N_HEADS*N_TRACKSIZE

;--------------
; Define a directory entry so we can extract the right parts

              RECORD    Attribs {
                        FA_Archive:1
                        FA_Directory:1
                        FA_Label:1
                        FA_System:1
                        FA_Hidden:1
                        FA_ReadOnly:1
                        }

              STRUC     DirEntry

FileName      DB        8 DUP (?)           ; name
FileExt       DB        3 DUP (?)           ; extension
FileAttrs     Attribs   ?                   ; attributes
              DB        10 DUP (?)          ; unused
FileTime      DW        ?                   ; time
FileDate      DW        ?                   ; date
FileStart     DW        ?                   ; starting cluster number
FileLength    DD        ?                   ; length in bytes

              ENDS      DirEntry


;-------------
; This does not start at offset 0100, so be careful about data accesses!
; Actual execution is at absolute address 0000:7C000...

              CODESEG

              STARTUPCODE

MainEntry:
              JMP       SHORT Booter        ; force two-byte jump
              NOP                           ; mandatory padding byte

;-------------
; Standard DOS 5.0 diskette boot record header

DiskParams    EQU       $

              DB        'Firmware'          ; 03 OEM name and ver (8 chars)
SectorSize    DW        512                 ; 0b bytes per sector
ClusterSize   DB        N_CLUSTERSIZE       ; 0d sectors per cluster
NumReserved   DW        1                   ; 0e reserved sectors
NumFATs       DB        2                   ; 10 number of FAT copies
MaxRootFiles  DW        N_ROOTFILES         ; 11 maximum root dir entries
NumSectors    DW        N_SECTORS           ; 13 total sectors on diskette
MediaID       DB        N_MEDIAID           ; 15 useless media descriptor
FATSize       DW        N_FATSIZE           ; 16 sectors in each FAT
TrackSize     DW        N_TRACKSIZE         ; 18 sectors per track (per head)
NumHeads      DW        N_HEADS             ; 1a number of heads
HiddenSectors DD        0                   ; 1c number of hidden sectors
              DD        0                   ; 20 number of sectors if > 32 MB
              DB        0                   ; 24 internal DOS drive ID
              DB        0                   ; 25 reserved
              DB        29h                 ; 26 Boot signature
              DD        000FF00EDh          ; 27 volume ID number
              DB        'Firmware386'       ; 2b volume label (11 chars)
              DB        'FAT12   '          ; 33 file-system type (8 chars)

;-------------
; Boot loader program

Booter:

;--- Move stack to just below our code load point
;    SP:SS = 0000:7C00, stack grows downward from there

              CLI                           ; for archaic 8086/8088 CPUs

              XOR       AX,AX               ; relocate stack
              MOV       SS,AX
              MOV       SP,7C00h            ;  ... just below our start

              STI

              MOV       ES,AX               ; set up ES as well

;--- Adjust DS so this code looks like it's at offset 100 in data
;    0000:7C00 = 07C0:0000 = 07B0:0100
;    This allows normal data addressing to work correctly without trickery

              MOV       AX,07B0h
              MOV       DS,AX

;--- need to modify diskette param block at 0:78 to match our params?

              NOP
              MOV       [ThisDrive],0       ; this is always true...

;--- locate and fetch first directory sector
;    this goes to 0000:7E00

              XOR       AX,AX               ; start with zeros
              MOV       AL,[NumFats]        ; account for sectors in both
              MUL       [FATSize]           ;  ... copies of the FATs
              ADD       AX,[WORD LOW HiddenSectors]   ; these are surely
              ADC       DX,[WORD HIGH HiddenSectors]  ;  ... unused
              ADD       AX,[NumReserved]    ; boot sector and other
              ADC       DX,0                ;  ... hidden sectors
              MOV       [AbsSector],AX      ; save start of dir

              CALL      CvtSectors          ; set up parameters
              MOV       BX,7E00h            ; set up target
              CALL      ReadSector          ; and get one sector

;--- skip volume labels and other debris
;    this also peels off Win 95 long names...

SkipNonFiles:
              TEST      [ES:BX+DirEntry.FileAttrs],(MASK FA_Label) OR \
                          (MASK FA_System) OR (MASK FA_Hidden)
              JZ        HaveFile
              ADD       BX,SIZE DirEntry
              JMP       SkipNonFiles
HaveFile:

;--- find sector address of data area just after directory
;    we use directory size in bytes rounded to next whole sector

              MOV       AX,SIZE DirEntry    ; entry size in bytes
              MUL       [MaxRootFiles]      ; times entries
              MOV       CX,[SectorSize]     ; round upward
              ADD       AX,CX               ; by sector size -  1
              DEC       AX
              DIV       CX                  ; get sectors
              ADD       [AbsSector],AX      ; add to dir start

;--- find file's starting sector from cluster number

              MOV       AX,[ES:BX+DirEntry.FileStart] ; first cluster
              SUB       AX,2                          ;  ... starts at 2
              MOV       CL,[ClusterSize]
              XOR       CH,CH
              MUL       CX                            ; AX = sectors
              ADD       [AbsSector],AX                ; save for later

;--- convert file length from bytes to sectors
;    we know the length must be < 64K bytes, so ignore high word

              MOV       AX,[WORD LOW ES:BX+DirEntry.FileLength]
              XOR       DX,DX               ; this is always 0
              DIV       [SectorSize]
              CMP       DX,0                ; any remainder?
              JZ        HaveLen
              INC       AX
HaveLen:
              MOV       [SectorCount],AX    ; set up for reading


;--- now pull in the file to 1000:0100
;    1000:0100 = physical address 10100, or just over 64K
;    This allows up to FFF0 bytes of data before a DMA boundary crossing

              MOV       BX,1000h            ; set up target address
              MOV       ES,BX
              MOV       BX,0100h

NextSector:
              MOV       AX,[AbsSector]      ; set up starting sector
              CALL      CvtSectors

              CALL      ReadSector          ; and get one sector

              INC       [AbsSector]         ; step to next secor
              ADD       BX,[SectorSize]     ;  ... and next target

              MOV       CX,[SectorCount]    ; continue until done
              DEC       [SectorCount]
              LOOP      NextSector

;--- set up regs for default *.COM program entry and do the deed
;    we put the stack at the far end of the segment, too

              MOV       AX,ES
              MOV       DS,AX

              CLI
              MOV       SS,AX
              MOV       SP,0fffeh
              STI

              MOV       BX,SP               ; set up zero at top of stack
              MOV       [WORD PTR ES:BX],0

              JMP       FAR 1000h:0100h      ; and we are off!


;-------------
; Convert from absolute sector number to cylinder/head/sector
; AX is the sector number (must be less than 64K for diskettes!)
; Returns result in globals

              PROC      CvtSectors

              XOR       DX,DX               ; MSW is always zero
              DIV       [TrackSize]         ; AX = tracks, DX = excess sects
              INC       DL                  ; sectors start at 1
              MOV       [ThisSector],DL     ; store result
              XOR       DX,DX               ; set up for division
              DIV       [NumHeads]          ; AX = cylinder, DX = head
              MOV       [ThisHead],DL       ; store results
              MOV       [ThisTrack],AX

              RET

              ENDP      CvtSectors


;-------------
; Read single sector using BIOS function
; Caller sets up ES:BX to target address

              PROC      ReadSector

              MOV       AX,[ThisTrack]      ; need to shuffle track bits
              MOV       CL,6                ;  ... get two MSB in high bits
              SHL       AH,CL
              OR        AH,[ThisSector]     ;  ... combine with sector
              MOV       CX,AX               ; now we can use CX
              XCHG      CH,CL               ; CH = track CL = sector + trk MSB
              MOV       DL,[ThisDrive]
              MOV       DH,[ThisHead]

              MOV       AX,0201h            ; read sectors, one at a time
              INT       13h

              JC        ShowError           ; lock up if failure

              RET

              ENDP      ReadSector



;-------------
; Display error indication on serial port...
; This is useful only if you've got an LED on the DTR line!
; We use the BIOS to meter out the blink rate for us

              PROC      ShowError

              MOV       DX,0040h            ; aim at BIOS data segment
              MOV       ES,DX

              MOV       BX,2*(COMPORT-1)    ; which port address?
              MOV       DX,[ES:BX]          ; get COM port base address
              ADD       DX,4                ; set modem control port offset

              PUSH      DX                  ; prepare for loop

ReShow:
              POP       DX                  ; recover port
              PUSH      DX
              MOV       AL,1
              OUT       DX,AL               ; DTR goes low

              MOV       CX,BLINK_OFF
              MOV       DX,0
              MOV       AH,86h
              INT       15h

              POP       DX                  ; recover port
              PUSH      DX
              MOV       AL,0
              OUT       DX,AL               ; DTR goes high

              MOV       CX,BLINK_ON
              MOV       DX,0
              MOV       AH,86h
              INT       15h

              JMP       ReShow              ; continue forever

              ENDP      ShowError

;-------------
; The little bit of data we need goes here

AbsSector     DW        ?                   ; absolute sector addresses

;--- used while reading diskette data

ThisSector    DB        ?                   ; sector number (1..N)
ThisHead      DB        ?                   ; head (0..N)
ThisTrack     DW        ?                   ; track (really cylinder, 0..N)
ThisDrive     DB        ?                   ; boot drive

SectorCount   DW        ?                   ; sectors remaining to read

;--- Pad to an even 512 bytes to fill the sector
;    Some BIOSes require the last two bytes to be 55 AA...
;     and refuse to boot if the sector doesn't pass that test

              ORG       02feh

              DB        055h
LastByte      DB        0aah

              END
