/* Micro-C program to exercise board addressing and buffering                */
/* Requires PORTIO.ASM library routines                                       */

#include <8086io.h>
#include "8086base.h"

#define BOARD_BASE   0x0300            /* board starting address             */
#define BOARD_SIZE   0x0020            /* address range on board             */
#define SYNC_ADDR    0x0378            /* address for non-selected I/O       */

#define LED_OFFSET   0x001E            /* offset of LED latches              */
#define SW_OFFSET    0x001E            /* offset of switch inputs            */

/*---------------------------------------------------------------------------*/
/* Delay for a specified number of BIOS timer ticks                          */
/* Returns current BIOS tick counter                                         */

WORD WaitTicks(Delay)
WORD Delay;
{

WORD CurrentTick;
WORD StopTick;

   StopTick = peekw(0x0040,0x006C) + Delay;

   do {
      CurrentTick = peekw(0x0040,0x006C);
   } while (CurrentTick != StopTick);

   return CurrentTick;

}

/*---------------------------------------------------------------------------*/
/* The main event                                                            */

char Option;
WORD Counter;
WORD PortOffset;
WORD PortValue;
WORD NextTick;

main() {

   serinit(9600,1);
   enable();

   putstr("I/O Logic Test Program\n");
   putstr("Embedded PC's ISA Bus Chapter 3 -- Ed Nisley\n");


   while (TRUE) {

      putstr("Board checkout routines:\n");
      putstr(" 1 - 74F521 board addr match (scope required)\n");
      putstr(" 2 - Address decoders (scope required)\n");
      putstr(" 3 - Write test data to LEDs\n");
      putstr(" 4 - Copy switches to LEDs\n");
      putstr(" 5 - Wait state calibration\n");
      putstr("Select option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '1' :
         putstr("Scope F521 pin 19 and -BIOW, verify hits and misses\n");
         printf("Board port base address is %04x, range is %04x\n",
                BOARD_BASE,BOARD_SIZE);
         Counter = 0;
         while (!chkch()) {
            for (PortOffset=0; PortOffset<BOARD_SIZE; PortOffset += 2) {
               outp(SYNC_ADDR,0x01);               /* sync on printer port   */
               outpw(BOARD_BASE+PortOffset,Counter);        /* board hit     */
               outp(SYNC_ADDR,0x00);
               ++Counter;
            }
         }
         break;
      case '2' :
         putstr("Scope LS138 and LS139 outputs, observe sequence\n");
         printf("Board port base address is %04x, range is %04x\n",
                BOARD_BASE,BOARD_SIZE);
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            for (PortOffset=0; PortOffset<BOARD_SIZE; PortOffset += 2) {
               outpw(BOARD_BASE+PortOffset,Counter);        /* write strobes */
               PortValue = inpw(BOARD_BASE+PortOffset);     /* read strobes  */
               ++Counter;
            }
            outp(SYNC_ADDR,0x00);
         }
         break;
      case '3' :
         putstr("Single bit chasing through LED\n");
         putstr("  segment order is DP, then ABCDEFG\n");
         do {
            outpw(BOARD_BASE+LED_OFFSET,0xFFFF);   /* all segments off       */
            WaitTicks(10);
            outpw(BOARD_BASE+LED_OFFSET,0x0000);   /* all segments on        */
            WaitTicks(10);
            outpw(BOARD_BASE+LED_OFFSET,0xFFFF);   /* all segments off       */
            WaitTicks(10);
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            for (Counter = 0x8000; Counter; Counter >>= 1) {
               outpw(BOARD_BASE+LED_OFFSET,~Counter);       /* bit low = on  */
               WaitTicks(7);
               Option = chkch();
               if (Option) {
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
            WaitTicks(10);
         } while (!Option);
         break;
      case '4' :
         putstr("Switch ON = blinking segment\n");
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            PortValue = inpw(BOARD_BASE+SW_OFFSET);         /* ON = low      */
            outpw(BOARD_BASE+LED_OFFSET,PortValue);         /* low = ON      */
            WaitTicks(1);
            outpw(BOARD_BASE+LED_OFFSET,0xFFFF);            /* all OFF       */
            WaitTicks(1);
            outp(SYNC_ADDR,0x00);
         }
         break;
      case '5' :
         putstr("Remove jumper from LS221 pin 3 to enable wait states\n");
         putstr("Scope LS221 pin 4 and -BIOW, tune pot for desired delay\n");
         putstr("Displayed count will drop at each additional wait state\n");
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            NextTick = WaitTicks(1) + 1;           /* lock on next edge      */
            Counter = 0;
            do {
               outpw(BOARD_BASE+LED_OFFSET,++Counter);
            } while (peekw(0x0040,0x006C) != NextTick);
            outp(SYNC_ADDR,0x00);
            printf("Count %5u\r",Counter);
         }
         putch('\n');
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
