; Generally useful utilities
; These are small enough to just include in the source
;  The Right Way is to produce a separate OBJ file, but...
;  with something this small, why bother?

;-------------
; Variables
; The assembler and linker sort the segments out correctly
;  even though everything is in one COM file with no actual segment info

              DATASEG

DataPort      DW        0                   ; at port base address (x78)
StatPort      DW        0                   ; base + 1             (x79)
CntlPort      DW        0                   ; base + 2             (x7A)


;-------------
; Set up LPT port addresses for later use

              CODESEG

              PROC      FindAddrs

              MOV       BX,2*(LPTPORT-1)+0008h        ; aim at port entry
              MOV       DX,[ES:BX]          ; get LPT port base address

              MOV       [DataPort],DX       ; save ports
              INC       DX
              MOV       [StatPort],DX
              INC       DX
              MOV       [CntlPort],DX

              RET

              ENDP      FindAddrs

;-------------
; Fetch control & status inputs into a single byte in AL
; Stomps on DX and AH
; Discards interrupt enable output bit and interrupt input
;  the output has no pin and the input would give us nine bits...
; Sets all four control outputs high to allow switch overrides

              CODESEG

              PROC      GetSwitches

;--- fetch status inputs

              MOV       DX,[StatPort]
              IN        AL,DX
              XOR       AL,LPT_StatInv      ; flip the inverted bits
              NOT       AL                  ; make closed switches into 1 bits

              RCL       AX,1                ; stash pin 11 in AH
              ROL       AL,1                ; discard pin 10 (interrupt)
              RCR       AX,1                ; recover pin 11 as MSB
              AND       AL,0f0h             ; strip off unused bits
              MOV       AH,AL               ; save for later

;--- fetch control inputs

              MOV       DX,[CntlPort]       ; aim at control port
              MOV       AL,NOT LPT_CntlInv  ; set outputs high
              AND       AL,NOT MASK PC_BiDi ; ensure we're not bidirectional!
              OUT       DX,AL
              Punt
              IN        AL,DX               ; read the inputs
              Punt
              XOR       AL,LPT_CntlInv      ; flip the inverted bits
              NOT       AL                  ; make closed switches into 1 bits

              AND       AL,00fh             ; strip off unused bits

; combine control and status bits into single return byte

              OR        AL,AH
              XOR       AH,AH               ; force high byte to zero

              RET

              ENDP      GetSwitches



;-------------
; Blinks bytes in AX on parallel port, does not return
; Presumes DataPort was set up by caller
; This is useful only if you've got LEDs on the data lines
; We use the BIOS to meter out the blink rate for us

ERR_TIME      EQU       2                   ; units of 64K microseconds

              PROC      ShowError

              PUSH      AX                  ; save bytes for loop

@@Again:
              POP       AX                  ; swap bytes
              XCHG      AH,AL
              PUSH      AX

              MOV       DX,[DataPort]       ; show low byte
              OUT       DX,AL

              MOV       CX,ERR_TIME
              MOV       DX,0
              MOV       AH,86h
              INT       15h

              JMP       @@Again             ; continue forever

              ENDP      ShowError



;-------------
; Determine what flavor of CPU and NDP we are running on
; Adapted from The Processor and Coprocessor, Robert Hummel, page 77
; AH = CPU type
;  01 - 8086/8
;  02 - 80286
;  03 - 80386DX or SX
;  03 - 80486DX or SX
; AL = NDP type
;  00 - none
;  01 - 8087
;  02 - 80287
;  03 - 80387DX or SX
;  04 - 80386DX or 80387SX
; Special case: return code 0400 = 80486SX w/o 80487SX

              UDATASEG
NDPStatus     DW        ?                   ; NDP status storage

              CODESEG
              P8086                         ; assume 16-bit code for now

              PROC      FindCpuID

              MOV       DX,0100h            ; set default return value

;--- 8086/8 always stack Flags bits 12-15 as 1

              PUSHF                         ; fetch the flags
              POP       AX

              AND       AH,00fh             ; clear high bits
              PUSH      AX
              POPF

              PUSHF                         ; fetch 'em again
              POP       AX

              AND       AH,0f0h             ; check high bits only
              CMP       AH,0f0h             ; if they're all on, then...
              JE        @@TestNDP           ;  this is an 8086/6

;--- 80286 always stacks those bits as zeros...

              INC       DH                  ; it's at least a 286

              PUSHF                         ; fetch the flags
              POP       AX

              OR        AH,0f0h             ; set high bits
              PUSH      AX
              POPF

              PUSHF                         ; fetch 'em again
              POP       AX

              AND       AH,0f0h             ; check high bits only
              JZ        @@TestNDP           ; if zero, it's an 80286

;--- the 386 doesn't have an alignment check bit

              P386                          ; want 32-bit instructions

              INC       DH                  ; it's at least a 386

              MOV       ECX,ESP             ; align stack to prevent error
              AND       ESP,NOT 3           ;  ... to a doubleword

              PUSHFD                        ; fetch the flags
              POP       EAX
              MOV       EBX,EAX             ; save for later

              XOR       EAX,00040000h       ; flip AC bit
              AND       EAX,NOT 00000200h   ; disable interrupts
              PUSH      EAX
              POPFD

              PUSHFD                        ; fetch 'em again
              POP       EAX

              XOR       EAX,EBX             ; did AC stay flipped?
              JZ        @@Restore           ;  no, must be '386

;--- the only thing left is a 486...

              INC       DH

;--- restore flags and stack pointer

@@Restore:
              PUSH      EBX
              POPFD

              MOV       ESP,ECX

;--- now check the NDP

              P8086                         ; want 16 bit code for now

@@TestNDP:
              MOV       [NDPStatus],-1      ; make this be nonzero for later

;--- first, does it respond with good status?

              FNINIT                        ; reset and initialize the chip
              FNSTSW    [NDPStatus]         ; fetch current status

              CMP       [NDPStatus],0       ; if nonzero, no NDP installed
              JNE       @@Done

;--- and a good control word?

              FNSTCW    [NDPStatus]         ; fetch control word
              AND       [NDPStatus],103fh   ; isolate valid bits
              CMP       [NDPStatus],003fh   ; does it look good?

              JNE       @@Done

;--- assume NDP matches the CPU type
;    verify that 386 doesn't have a 287

              MOV       DL,DH

              CMP       DH,3                ; if not 386, we're done
              JNE       @@Done

;--- 287 and 387 have different representations for +/- infinity

              FLD1                          ; stack a 1
              FLDZ                          ; stack a 0
              FDIV                          ; get an infinity by division

              FLD       ST                  ; duplicate it
              FCHS                          ; flip the sign
              FCOMPP                        ; compare and discard

              FSTSW     [NDPStatus]         ; store the result
              MOV       AX,[NDPStatus]      ; fetch it
              SAHF                          ;  ... to the flags

              JNE       @@Done              ; if not equal, it's a 387
              DEC       DL                  ;  ... otherwise, 287

;-- finished!

@@Done:
              MOV       AX,DX               ; set up return value

              RET

              ENDP      FindCpuID
