/* Embedded PC's ISA Bus Chapter 4 -- Ed Nisley                              */
/* Micro-C program to exercise 8254 timer                                    */
/* Requires PORTIO.ASM library routine                                       */

#include <8086io.h>
#include "8086base.h"

#define BOARD_BASE   0x0300            /* board starting address             */
#define BOARD_SIZE   0x0020            /* address range on board             */
#define SYNC_ADDR    0x0378            /* address for non-selected I/O       */

#define LED_OFFSET   0x001E            /* offset of LED latches              */
#define SW_OFFSET    0x001E            /* offset of switch inputs            */

#define I8254_BASE   (BOARD_BASE+0x0008)           /* base of timer regs     */

#define HI(x) (x >> 8)                 /* move high byte to low byte         */
#define LO(x) (x & 0x00FF)             /* isolate low byte                   */

#define LOADTIMER(i,c,t) outpw(I8254_BASE+6,c); \
                         outpw(I8254_BASE+2*i,t & 0x00ff); \
                         outpw(I8254_BASE+2*i,t >> 8);

#define LOADTIMERB(i,c,t) outp(I8254_BASE+6,c); \
                          outp(I8254_BASE+2*i,t & 0x00ff); \
                          outp(I8254_BASE+2*i,t >> 8);

/*---------------------------------------------------------------------------*/
/* Set up BIOS real-time alarm clock to delay interval in microseconds       */
/* Returns 0 if OK, error code if not (also sets *pbAlarm to return code)    */
/* High-order bit of *pbAlarm is set when delay time expires                 */
/* Remember to split the microsecond count manually... 1 sec = 1000000 us    */
/* The asm{} section starts with BP pushed and loaded with SP                */

WORD SetAlarm(DelayHi,DelayLow,pbAlarm)
WORD DelayHi;
WORD DelayLow;
BYTE *pbAlarm;
{

   DelayHi;                   /* keep compiler happy                         */
   DelayLow;
   pbAlarm;

asm {
   MOV   BX,SP                aim at the stack
   MOV   CX,8[BX]             get delay high byte
   MOV   DX,6[BX]               ... low byte
   MOV   BX,4[BX]             get pointer to alarm byte
   MOV   AX,DS                  ... set up segment
   MOV   ES,AX
   MOV   AX,#$8300            Set Event Wait Interval function
   INT   $15
   JC    ?seterr              C set on error
   XOR   AL,AL                  ... clear return flag
?seterr  MOV   BX,SP          set up for return code
   MOV   BX,4[BX]
   MOV   ES:[BX],AL
   XOR   AH,AH                clear high byte of return value
}

}


/*---------------------------------------------------------------------------*/
/* Convert byte into two seven-segment digits                                */
/* Returns 16-bit value with both decimal points turned off                  */
/*  ... remember to invert the value before writing it out                   */

WORD SegTable[16] = {0x007E,0x0030,0x006D,0x0079,
                     0x0033,0x005B,0x005F,0x0070,
                     0x007F,0x007B,0x0077,0x001F,
                     0x004E,0x003D,0x004F,0x0047};

WORD ByteToSegs(Value)
int Value;
{

WORD RetValue;

   Value &= 0x00FF;                    /* make sure it is a byte             */

   RetValue = (SegTable[Value >> 4] << 8) | SegTable[Value & 0x000F];

   return RetValue;


}


/*---------------------------------------------------------------------------*/
/* The main event                                                            */

char Option;
WORD Counter;
WORD PortOffset;
WORD PortValue;
BYTE Alarm;
WORD Status[3];
WORD Edges[3];
int  Timer;

main() {

   serinit(9600,1);
   enable();

   putstr("\nHigh-resolution Timer Test Program\n");
   putstr("Embedded PC's ISA Bus Chapter 4 -- Ed Nisley\n");

   while (TRUE) {

      putstr("Timer checkout routines:\n");
      putstr(" 1 - Read/write all timer addresses with test pattern\n");
      putstr(" 2 - Set up counters for scope observation\n");
      putstr(" 3 - Start and read back counters, report results\n");
      putstr(" 4 - Same as 3, but use byte I/O\n");
      putstr("Select option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '1' :
         putstr("Scope 8254 pin 23 to verify -WR, pin 22 for -RD,\n");
         putstr("  pin 21 is -CS and should match each WR/RD to the 8254\n");
         printf("8254 base address is %04x\n", I8254_BASE);
         putstr("Uses 16-bit I/O operations\n");
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x02);                  /* sync at top of loop    */
            for (PortOffset=0; PortOffset<8; PortOffset += 2) {
               outp(SYNC_ADDR,0x01);               /* sync on printer port   */
               outpw(I8254_BASE+PortOffset,Counter);        /* chip hit      */
               inpw(I8254_BASE+PortOffset);                 /* chip hit      */
               outp(SYNC_ADDR,0x00);
               ++Counter;
            }
         }
         break;
      case '2' :
         putstr("Timer 0 = 1 ms square wave @ pin 10\n");
         putstr("Timer 1 = 2 ms square wave @ pin 13\n");
         putstr("Timer 2 = 3 ms square wave @ pin 17\n");
         putstr("Uses 8-bit I/O operations\n");
         outp(SYNC_ADDR,0x01);                     /* sync at start          */
         LOADTIMERB(0,0x0036,7159);                /*  1 ms @ 7.1591 MHz     */
         LOADTIMERB(1,0x0076,14318);               /*  2 ms @ 7.1591 MHz     */
         LOADTIMERB(2,0x00B6,21477);               /*  3 ms @ 7.1591 MHz     */
         outp(SYNC_ADDR,0x00);
         putstr(" ... timers are now running...\n\n\n");
         break;
      case '3' :
         putstr("Timers should produce 1000, 500, and 333 counts/sec\n");
         putstr("  +/- one or two counts for sampling error...\n");
         putstr("LEDs display iteration counter in hex\n");
         putstr("Uses 16-bit I/O operations\n");
         outp(SYNC_ADDR,0x02);                     /* sync at start          */
         LOADTIMER(0,0x0036,7159);                 /*  1 ms @ 7.1591 MHz     */
         LOADTIMER(1,0x0076,14318);                /*  2 ms @ 7.1591 MHz     */
         LOADTIMER(2,0x00B6,21477);                /*  3 ms @ 7.1591 MHz     */
         Counter = 0;
         while (!chkch()) {
            Status[0] = Status[1] = Status[2] = 0;
            Edges[0] = Edges[1] = Edges[2] = 0;
            if (SetAlarm(0x000F,0x4240,&Alarm)) {  /* alarm in 1 second      */
               printf("Cannot use BIOS alarm function, code %02x\n",Alarm);
               break;
            }
            outp(SYNC_ADDR,0x01);
            while (!Alarm) {
               outpw(I8254_BASE+6,0x00EE);         /* latch all status bytes */
               for (Timer=0; Timer<3; ++Timer) {
                  PortValue = inpw(I8254_BASE+2*Timer) & 0x0080;
                  Edges[Timer] += (0 != (Status[Timer] ^ PortValue));
                  Status[Timer] = PortValue;       /* save for next pass     */
               }
            }
            outp(SYNC_ADDR,0x00);
            printf("\r%5u %5u %5u counts/sec",
                   Edges[0]/2,Edges[1]/2,Edges[2]/2);
            outpw(BOARD_BASE+LED_OFFSET,~ByteToSegs(Counter));
            ++Counter;
         }
         putstr("Counters continue to run...\n");
         break;
      case '4' :
         putstr("Timers should produce 1000, 500, and 333 counts/sec\n");
         putstr("  +/- one or two counts for sampling error...\n");
         putstr("LEDs display iteration counter in hex\n");
         putstr("Uses 8-bit I/O operations\n");
         outp(SYNC_ADDR,0x02);                     /* sync at start          */
         LOADTIMERB(0,0x0036,7159);                /*  1 ms @ 7.1591 MHz     */
         LOADTIMERB(1,0x0076,14318);               /*  2 ms @ 7.1591 MHz     */
         LOADTIMERB(2,0x00B6,21477);               /*  3 ms @ 7.1591 MHz     */
         Counter = 0;
         while (!chkch()) {
            Status[0] = Status[1] = Status[2] = 0;
            Edges[0] = Edges[1] = Edges[2] = 0;
            if (SetAlarm(0x000F,0x4240,&Alarm)) {  /* alarm in 1 second      */
               printf("Cannot use BIOS alarm function, code %02x\n",Alarm);
               break;
            }
            outp(SYNC_ADDR,0x01);
            while (!Alarm) {
               outp(I8254_BASE+6,0x00EE);          /* latch all status bytes */
               for (Timer=0; Timer<3; ++Timer) {
                  PortValue = inp(I8254_BASE+2*Timer) & 0x0080;
                  Edges[Timer] += (0 != (Status[Timer] ^ PortValue));
                  Status[Timer] = PortValue;       /* save for next pass     */
               }
            }
            outp(SYNC_ADDR,0x00);
            printf("\r%5u %5u %5u counts/sec",
                   Edges[0]/2,Edges[1]/2,Edges[2]/2);
            outpw(BOARD_BASE+LED_OFFSET,~ByteToSegs(Counter));
            ++Counter;
         }
         putstr("Counters continue to run...\n");
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
