/* Micro-C program to demonstrate interrupt handler peculiarities            */
/* Requires PORTIO.ASM and FIRMDEV.ASM library routines                      */
/* Presumes your PC terminal emulator can handle ANSI cursor controls...     */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

#define NEST_1xxx                      /* allows nested interrupts           */
#define SYNCH                          /* ensures timers are synchronized    */

#define STACKSIZE 1000                 /* maximum stack length in bytes      */

/*---------------------------------------------------------------------------*/
/* Globals                                                                   */

WORD  Int08Seg;                        /* stored BIOS timer tick vector      */
WORD  Int08Off;

WORD  Int70Seg;                        /* stored BIOS alarm delay vector     */
WORD  Int70Off;

/*--- array holding response times and calculated values                     */
/*    the LOAD and NOW values are unsigned, the rest are < 32K               */

#define RESP_LOAD 0
#define RESP_NOW  1
#define RESP_SET  2
#define RESP_SAVE 3
#define RESP_MIN  4
#define RESP_AVG  5
#define RESP_MAX  6

int Response[3][7];

/*---------------------------------------------------------------------------*/
/* Timer 0 hardware interrupt handler                                        */
/* Assumes interrupt on primary controller                                   */

INTERRUPT(Timer0Handler) HandlerT0() {

   outp(SYNC_ADDR,inp(SYNC_ADDR) | 0x01);

   if (!(Response[0][RESP_SET]++)) {
      Response[0][RESP_NOW] = ReadTimer(0,I8254_BASE);      /* fetch times   */
   }

   outp(I8259A,NS_EOI);

   outp(SYNC_ADDR,inp(SYNC_ADDR) & ~0x01);

   return;

}


/*---------------------------------------------------------------------------*/
/* Timer 1 hardware interrupt handler                                        */
/* Assumes interrupt on secondary controller                                 */

INTERRUPT(Timer1Handler) HandlerT1() {

   outp(SYNC_ADDR,inp(SYNC_ADDR) | 0x02);

#ifdef NEST_1

   outp(I8259B,NS_EOI);                /* tell slave we are done             */
   outp(I8259A,NS_EOI);                /* tell master we are done            */
   enable();                           /* and allow other interrupts         */

#endif

   if (!(Response[1][RESP_SET]++)) {
      Response[1][RESP_NOW] = ReadTimer(1,I8254_BASE);      /* fetch times   */
   }

#ifndef NEST_1

   outp(I8259B,NS_EOI);                /* tell slave we are done             */
   outp(I8259A,NS_EOI);                /* tell master we are done            */

#endif

   outp(SYNC_ADDR,inp(SYNC_ADDR) & ~0x02);

   return;

}




/*---------------------------------------------------------------------------*/
/* Timer 2 hardware interrupt handler                                        */
/* Assumes interrupt on secondary controller                                 */
/* CAUTION -- all the constants are hardcoded in here...                     */
/*  ... and the registers are saved manually                                 */
/* Micro-C inserts PUSH BP and MOV BP,SP before the first line of code       */

HandlerT2() {

asm {
*
   PUSH  AX                            save bystanders
   PUSH  DX
*
   MOV   AL,#$80                       latch Timer 2
   MOV   DX,#$030E
   OUT   DX,AL
*
   MOV   DX,#$0378                     flag startup
   IN    AL,DX
   OR    AL,#$04
   OUT   DX,AL
*
   MOV   AX,CS                         set up data segment addressing
   MOV   DS,AX
*
   PUSH  SI                            save more bystanders
   PUSH  DS
*
   MOV   DX,#$030C                     read the latched LSB
   IN    AL,DX
   JMP   <Punt4
Punt4    NOP
   MOV   AH,AL
   IN    AL,DX                          ... and the MSB
   XCHG  AH,AL                         rearrange them
   MOV   DX,AX                         save for later
*
   MOV   SI,#Response+(2*7*2)          aim at our slice of Response
*
   MOV   AX,4[SI]                      fetch _SET element
   AND   AX,AX                         previous entry processed?
   JNZ   Done                          nonzero says skip this one
*
   MOV   2[SI],DX                      stash it away
   INC   ASWORD(4[SI])                 account for this sample
*
Done     MOV   AL,#$20                 send the EOIs
   OUT   $20,AL                         ... to slave
   OUT   $A0,AL                         ... and master
*
   MOV   DX,#$0378                     flag shutdown
   IN    AL,DX
   AND   AL,#$FB
   OUT   DX,AL
*
   POP   DS                            restore bystanders
   POP   SI
   POP   DX
   POP   AX
*
   POP   BP                            saved at function entry
   IRET                                preempt normal return
}

}


/*---------------------------------------------------------------------------*/
/* BIOS timer tick hook                                                      */
/* Wraps an output sync around the timer tick to show collisions             */
/* Micro-C inserts PUSH BP and MOV BP,SP before the first line of code       */

Int08Hook() {

asm {
*
   PUSH  AX                            save bystanders
   PUSH  DX
*
   MOV   DX,#$0378                     flag startup
   IN    AL,DX
   OR    AL,#$08
   OUT   DX,AL
*
   INT   $60                           pass control to the BIOS code
*
   MOV   DX,#$0378                     flag shutdown
   IN    AL,DX
   AND   AL,#$F7
   OUT   DX,AL
*
   POP   DX                            restore bystanders
   POP   AX
*
   POP   BP
   IRET                                restore stacked flags
}

}

/*---------------------------------------------------------------------------*/
/* BIOS alarm delay hook                                                     */
/* Wraps an output sync around the real-time clock tick to show collisions   */
/* Micro-C inserts PUSH BP and MOV BP,SP before the first line of code       */

Int70Hook() {

asm {
*
   PUSH  AX                            save bystanders
   PUSH  DX
*
   MOV   DX,#$0378                     flag startup
   IN    AL,DX
   OR    AL,#$10
   OUT   DX,AL
*
   INT   $61                           pass control to the BIOS code
*
   MOV   DX,#$0378                     flag shutdown
   IN    AL,DX
   AND   AL,#$EF
   OUT   DX,AL
*
   POP   DX                            restore bystanders
   POP   AX
*
   POP   BP
   IRET                                restore stacked flags
}

}



/*---------------------------------------------------------------------------*/
/* Check stack usage                                                         */
/* We're positive that the search will terminate before the max length!      */
/* Also presumes the stack starts at the end of the segment...               */
/*  which is true by definition for files loaded by the diskette booter      */

WORD CheckStack() {

asm {
   MOV   AX,SS                         search in stack segment
   MOV   ES,AX                          ... set it up

   MOV   CX,#STACKSIZE/2               set up search length in words
   MOV   DI,#-STACKSIZE                from lowest stack address
   XOR   AX,AX                         searching for non-zeros
   REPE                                go forth and do...
   SCASW

   ADD   CX,CX                         convert back to bytes
   MOV   AX,CX                         return remaining count as length

}

   return;

}



/*---------------------------------------------------------------------------*/
/* The main event                                                            */

main() {

WORD Alarm;
int  Timer;
WORD TimerValue;
int  Counter;

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("Interrupt Handler Timing Exerciser\n");
   putstr("Embedded PC's ISA Bus Chapter 5 -- Ed Nisley\n");

#ifdef NEST_1
   putstr("HandlerT1 is running in nested interrupt mode\n");
#endif
#ifndef SYNCH
   putstr("Timers are not synchronized\n");
#endif

   if (SetAlarm(0x000F,0x4240,&Alarm)) {  /* alarm in 1 second      */
      printf("Cannot use BIOS alarm function, code %02x\n",Alarm);
Stall: goto Stall;
   }
   putstr("Initial alarm time set successfully\n");

   putstr("Preloading response array... ");
   for (Timer=0; Timer<3; ++Timer) {
      Response[Timer][RESP_MIN] = 32767;
#ifdef SYNCH
      Response[Timer][RESP_LOAD] = 0x8c73;         /* 5 ms with no offset    */
#else
      Response[Timer][RESP_LOAD] = 0x8c73 + Timer; /* 5 ms + slight offset   */
#endif
   }
   putstr("done\n");

   putstr("Setting interrupt vectors... ");
   SetVector(I8259A_VECTOR+5,GetCodeSeg(),INT_ENTRY(HandlerT0));
   SetVector(I8259B_VECTOR+2,GetCodeSeg(),INT_ENTRY(HandlerT1));
   SetVector(I8259B_VECTOR+7,GetCodeSeg(),&HandlerT2);

   GetVector(I8259A_VECTOR+0,&Int08Seg,&Int08Off); /* capture sys timer      */
   SetVector(0x60,Int08Seg,Int08Off);              /* move to Int 60         */
   SetVector(I8259A_VECTOR+0,GetCodeSeg(),&Int08Hook);      /* install ours  */

   GetVector(0x70,&Int70Seg,&Int70Off);            /* capture timer tick     */
   SetVector(0x61,Int70Seg,Int70Off);              /* move to Int 61         */
   SetVector(0x70,GetCodeSeg(),&Int70Hook);        /* install our handler    */

   putstr("done\n");

   putstr("Starting timers... ");
   for (Timer=0; Timer<3; ++Timer) {
      LoadTimer(Timer,0x34,Response[Timer][RESP_LOAD],I8254_BASE);
   }
   putstr("running\n");

   putstr("Enabling interrupts... ");
   disable();
   outp(I8259A+1,inp(I8259A+1) & ~0x20);           /* IRQ5 enabled           */
   outp(I8259B+1,inp(I8259B+1) & ~0x04);           /* IRQ10 enabled          */
   outp(I8259B+1,inp(I8259B+1) & ~0x80);           /* IRQ15 enabled          */
   enable();
   putstr(" active\n");

   putstr("\nInterrupt response times in 139 ns ticks:\n");
   putstr("Timer Current Minimum Average Maximum\n");
   putstr("\n\n\n\n");                 /* make sure we have room             */
   putstr("Press any key to clear min/max values\r");
   putstr("\x1b[4A");                  /* up to first blank                  */
   putstr("\x1b[s");                   /* save cursor position               */

   Counter = 0;
   while (TRUE) {

      outp(SYNC_ADDR,0x80);

/*--- update statistics for each timer                                       */
/*    the timers count down, so the first step converts to "elapsed steps"   */

      for (Timer=0; Timer<3; ++Timer) {
         if (Response[Timer][RESP_SET]) {
            outp(SYNC_ADDR,0x40);
            TimerValue = Response[Timer][RESP_LOAD] -
                         Response[Timer][RESP_NOW];
            Response[Timer][RESP_SAVE] = TimerValue;
            Response[Timer][RESP_MIN] = min(Response[Timer][RESP_MIN],
                                            TimerValue);
            Response[Timer][RESP_MAX] = max(Response[Timer][RESP_MAX],
                                            TimerValue);
            Response[Timer][RESP_AVG] = (299*Response[Timer][RESP_AVG] +
                                         TimerValue) / 300;
            Response[Timer][RESP_SET] = 0;
            outp(SYNC_ADDR,0x00);
         }
      }

/*--- dump statistics when alarm goes off (once a second)                    */

      if (Alarm) {
         outp(SYNC_ADDR,0x20);
         for (Timer=0; Timer<3; ++Timer) {
            printf("%3u    %6d  %6d  %6d  %6d\n",Timer,
               Response[Timer][RESP_SAVE],
               Response[Timer][RESP_MIN],
               Response[Timer][RESP_AVG],
               Response[Timer][RESP_MAX]);
         }
         printf("Max stack used: %04x",CheckStack());
         putstr("\x1B[u");             /* restore cursor position            */
         putstr("\x1b[s");             /* save cursor position               */
         Alarm = 0;
         SetAlarm(0x000F,0x4240,&Alarm);           /* reset alarm            */
         outpw(LED_ADDR,~ByteToSegs(++Counter));   /* show running count     */
         outpw(SYNC_ADDR,0x00);
      }

/*--- check for keypress to reset the min/max values                         */

      if (chkch()) {
         for (Timer=0; Timer<3; ++Timer) {
            Response[Timer][RESP_MIN] = 32767;
            Response[Timer][RESP_MAX] = 0;
         }
      }

      outp(SYNC_ADDR,0x00);

   }


}
