/* Embedded PC's ISA Bus Chapter 6 -- Ed Nisley                              */
/* Micro-C program to exercise Firmware Development Board memory             */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

#define MAX_ERRORS   10          /* stop testing after this many mismatches  */

/*---------------------------------------------------------------------------*/
/* The main event                                                            */

char Option;
WORD Counter;
WORD RAMSeg;
WORD RAMAddr;
WORD RAMSize;
BYTE TestData;
int  ErrorCounter;

BYTE PRSData;
BYTE *pPRSData;
WORD PRSSegment;

BYTE Alarm;

main() {

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("\nNon-volatile memory tests\n");
   putstr("Embedded PC's ISA Bus Chapter 6 -- Ed Nisley\n");

   putstr("Calculating pseudorandom data...  ");
asm {
   MOV   >RANDSEED,#$0001
}
   PRSSegment = GetDataSeg() + 0x1000;             /* step 64K above data    */
   for (pPRSData = 0; pPRSData != 0x8000; pPRSData += 2) {
      pokew(PRSSegment,pPRSData,rand(0x8000));
   }
   putstr("done\n");

   while (TRUE) {

      putstr("\n\nMemory checkout routines:\n");
      putstr(" 1 - Verify address match comparator\n");
      putstr(" 2 - Repeated memory reads\n");
      putstr(" 3 - Repeated memory writes (=NOT= for EEPROM!)\n");
      putstr(" 4 - Load 8K EEPROM/RAM w/PRS8K data (check w/#5)\n");
      putstr(" 5 - Verify 8K PRS8K data\n");
      putstr(" 6 - Verify 32K PRS32K data\n");
      putstr("Select option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '1' :
         putstr("Sync on LPT bit 0 & 1, scope RAM select to verify hits\n");
         printf("Memory segment is %04x\n",NV_SEGMENT);
         Counter = 0;
         RAMSeg = NV_SEGMENT & 0xF000;             /* strip off low bits     */
         RAMAddr = 0x0000;
         while (!chkch()) {
            if (!RAMAddr) {
               outp(SYNC_ADDR,0x01);               /* sync on printer port   */
            }
            outp(SYNC_ADDR,0x02);                  /* mark each peek         */
            TestData = peek(RAMSeg,RAMAddr);                /* pick it up    */
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(HI(Counter)));
            ++Counter;
            RAMAddr += 0x1000;
         }
         break;
      case '2' :
         RAMSize = 0x8000;
         RAMSeg = NV_SEGMENT;                      /* set up for assembler   */
         printf("Reading memory at segment %04x, size %04x\n",
                NV_SEGMENT,RAMSize);
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
asm {
   MOV   CX,RAMSize        repeat for whole block
   PUSH  DS
   MOV   AX,RAMSeg         set up EPROM address
   MOV   DS,AX
   XOR   SI,SI
   REP
   LODSB
   POP   DS
}
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(Counter));
            ++Counter;
         }
         break;
      case '3' :
         printf("Writing memory at segment %04x, size %04x\n",
                NV_SEGMENT,NV_SIZE);
         RAMAddr = 0x0000;
         Counter = 0;
         while (!chkch()) {
            if (RAMAddr == NV_SIZE) {
               outp(SYNC_ADDR,0x01);               /* sync on printer port   */
               outpw(LED_ADDR,~ByteToSegs(Counter));
               ++Counter;
               RAMAddr = 0;
            }
            outp(SYNC_ADDR,0x02);                  /* mark each peek         */
            poke(NV_SEGMENT,RAMAddr,~RAMAddr);     /* write inverse address  */
            outp(SYNC_ADDR,0x00);
            ++RAMAddr;
         }
         break;
      case '4' :
         printf("Loading 8K EEPROM at seg %04x\n",NV_SEGMENT);
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != 0x2000; ++RAMAddr) {
            PRSData = peek(PRSSegment,RAMAddr);
            outp(SYNC_ADDR,0x01);                  /* mark the write         */
            poke(NV_SEGMENT,RAMAddr,PRSData);      /* start it               */
            if (SetAlarm(0x0000,0x2710,&Alarm)) {  /* error after 10 ms      */
               putstr("Cannot set BIOS timeout!\n");
               break;
            }
            do {
               TestData = peek(NV_SEGMENT,RAMAddr);
            } while ((0x0080 & (PRSData ^ TestData)) && !Alarm);
            if (Alarm) {
               printf("Timout programming %04x to %02x, is %02x\n",
                  RAMAddr,PRSData,TestData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            TestData = peek(NV_SEGMENT,RAMAddr);
            if (TestData != PRSData) {
               printf("Cannot program %04x to %02x, is %02x\n",
                  RAMAddr,PRSData,TestData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            if (CancelAlarm(0x0000,0x2710,&Alarm)) {
               putstr("Cannot cancel BIOS timeout!\n");
               break;
            }
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(HI(RAMAddr)));
         }
         putstr("... done\n");
         break;
      case '5' :
         printf("Verifying 8K block at seg %04x\n",NV_SEGMENT);
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != 0x2000; ++RAMAddr) {
            outp(SYNC_ADDR,0x01);                  /* mark this loop         */
            TestData = peek(NV_SEGMENT,RAMAddr);
            PRSData = peek(PRSSegment,RAMAddr);
            if (PRSData != TestData) {
               printf("Mismatch at %04x, data is %02x, should be %02x\n",
                       RAMAddr,TestData,PRSData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
         }
         putstr("... done\n");
         break;
      case '6' :
         printf("Verifying 32K block at seg %04x\n",NV_SEGMENT);
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != 0x8000; ++RAMAddr) {
            outp(SYNC_ADDR,0x01);                  /* mark this loop         */
            TestData = peek(NV_SEGMENT,RAMAddr);
            PRSData = peek(PRSSegment,RAMAddr);
            if (PRSData != TestData) {
               printf("Mismatch at %04x, data is %02x, should be %02x\n",
                       RAMAddr,TestData,PRSData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
         }
         putstr("... done\n");
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
