/* Embedded PC'S ISA Bus Chapter 7 -- Ed Nisley                              */
/* Micro-C program to exercise Firmware Development Board watchdog chip      */
/* This =must= be booted from diskette to get control before the first reset */
/*  (you can leave the reset connector off until you see how this works...)  */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"
#include "macros.inc"

#define WATCH_RATE   3                 /* BIOS timer ticks per watchdog bit  */

/*---------------------------------------------------------------------------*/
/* Globals                                                                   */

WORD  Int02Off;                        /* old NMI handler vector             */
WORD  Int02Seg;

WORD  Int1COff;                        /* old BIOS post-tick vector          */
WORD  Int1CSeg;

WORD  WatchBits;                       /* bits for next shift register load  */
WORD  WatchShift;                      /* watchdog output shift register     */
WORD  WatchCounter;                    /* counts shifts                      */
WORD  WatchPending;                    /* nonzero when new bits needed       */
BYTE  WatchDivide;                     /* divides BIOS tick rate             */

WORD  CtlsCopy;                        /* copy of port bits                  */

/*---------------------------------------------------------------------------*/
/* Non-maskable interrupt handler                                            */
/* This tests for a power-fail warning interrupt                             */
/* You will see a single decimal point while this code is active             */
/* Micro-C inserts PUSH BP and MOV BP,SP before the first line of code       */

HandlerNMI() {

asm {
*
         PUSH     AX                   save bystanders
         PUSH     DX
         PUSH     DS
         MOV      AX,CS                aim at our segment again
         MOV      DS,AX
*
* Check to see if the power fail bit is active
*
         MOV      DX,#STAT_ADDR_A
         IN       AX,DX
         Punt
         TEST     AX,#PWR_GOOD_A
         JNZ      NMI_Chain            nonzero says not our problem
*
* We have a power failure, so write-protect the RAM and lock up
*
         MOV      DX,#CTLS_ADDR_A      turn off the write-enable bit
         MOV      AX,#~NV_WENABLE_A
         OUT      DX,AX
         Punt
*
         MOV      DX,#LED_ADDR_A       show that we are locking up
         MOV      AX,#~$0080            ... with right decimal point ON
         OUT      DX,AX
         Punt
*
NMI_Lock JMP      <NMI_Lock            jam up here until next reset
*
*
* Chain to previous NMI handler
*
NMI_Chain POP     DS                   restore bystanders
         POP      DX
         POP      AX
         POP      BP
         JMP      CS:0:>Int02Off       indirect to old handler through CS
}

}


/*---------------------------------------------------------------------------*/
/* Watchdog driver                                                           */
/* This shifts a 16-bit value out through the watchdog pin                   */
/* We capture the BIOS timer post-tick interrupt to time the output          */
/* Mainline code must reload the shift value on demand to prevent a timeout  */
/* Micro-C inserts PUSH BP and MOV BP,SP before the first line of code       */

HandlerWD() {

asm {
*
         PUSH     AX                   save bystanders
         PUSH     DX
         PUSH     DS
         MOV      AX,CS                aim at our segment again
         MOV      DS,AX
*
* Count down the interrupts until we need a watchdog update
*
         DEC      <WatchDivide
         JNZ      WD_Ret
         MOV      <WatchDivide,#WATCH_RATE
*
* Decide if a new watchdog word is needed
* If it is, and the mainline code is jammed, we lock up and die
*
WD_Tick  DEC      >WatchCounter
         JNZ      WD_Go                nonzero says use old bits
         CMP      >WatchPending,#0     has mainline code reloaded the bits?
         JE       WD_Load              zero says yes, so we are golden
         MOV      DX,#LED_ADDR_A       nonzero says we have trouble
         MOV      AX,#~$8000           left decimal point flags the problem
         OUT      DX,AX
WD_Lock  JMP      <WD_Lock             stay here until watchdog timeout
*
WD_Load  MOV      AX,WatchBits         fetch new bits
         MOV      WatchShift,AX         ... for the shift reg
         MOV      >WatchCounter,#16    reload the counter
         INC      >WatchPending        set flag for mainline code
*
* Blip the watchdog output to ensure a transition every time
*
WD_Go    MOV      DX,#CTLS_ADDR_A      set up for watchdog output
         MOV      AX,CtlsCopy          get existing bits
         AND      AX,#~WATCHDOG_A      send a low (LED ON)
         OUT      DX,AX
         Punt
         OR       AX,#WATCHDOG_A       send a high (LED OFF)
         OUT      DX,AX
*
* Rotate the watchdog bits and send the high one
* We flip the bit so 1 turns the LED ON like it should
*
         ROL      >WatchShift,1        get high-order bit in C
         JNC      WD_Z                 clear says leave the output high
         AND      AX,#~WATCHDOG_A      set says make output low
WD_Z     OUT      DX,AX                send it out
         XOR      AX,#WATCHDOG_A       flip the bit back again
         MOV      CtlsCopy,AX          save for next time
*
*
WD_Ret   POP      DS                   restore bystanders
         POP      DX
         POP      AX
*
         POP      BP
         IRET                          restore stacked flags
}

}


/*---------------------------------------------------------------------------*/
/* The main event                                                            */

main() {

WORD PortValue;
int  Counter;
char Option;
int  KillWatch;

   CtlsCopy = 0;                       /* set initial bits & fast watchdog   */
   outpw(CTLS_ADDR,CtlsCopy);

   WatchBits = WatchPending = 0;
   WatchShift = 0x80F0;
   WatchDivide = WATCH_RATE;                       /* 2 * 55 ms per shift    */
   WatchCounter = 1;                               /* reload on first tick   */

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("Watchdog timer and power monitor tests\n");
   putstr("Embedded PC's ISA Bus Chapter 7 -- Ed Nisley\n");

   putstr("Setting interrupt vectors... ");
   GetVector(0x02,&Int02Seg,&Int02Off);            /* non-maskable interrupt */
   SetVector(0x02,GetCodeSeg(),&HandlerNMI);

   GetVector(0x1C,&Int1CSeg,&Int1COff);            /* BIOS timer post-tick   */
   SetVector(0x1C,GetCodeSeg(),&HandlerWD);
   putstr("done\n");

   putstr("Enabling I/O channel check interrupt... ");
   PortValue = inp(SYS_CTLS);
   outp(SYS_CTLS,PortValue & 0x0037);              /* bit 3 = 0 to enable    */
   putstr("done\n");

   putstr("\nAdjust PFI pot to trigger NMI, right decimal goes ON\n");
   putstr("Press 1 to force watchdog timeout, left decimal goes on\n");

   Counter = 0;
   KillWatch = 0;
   while (TRUE) {

      Option = chkch();

      if (WatchPending && !KillWatch) {            /* need new bits?         */
         WatchBits = ~inpw(SW_ADDR);               /* fetch new switches     */
         WatchPending = 0;
         printf("%04x\r",Counter);                 /* show some activity     */
      }

      switch (Option) {
      case '1' :
         putstr("Watchdog reload suppressed\n");
         KillWatch = 1;
         break;
      case 0 :
         break;
      default :
         putstr("Invalid option...\n");
      }

      BIOSWait(0x0001,0x86a0);                     /* punt for 100 ms        */

      outpw(LED_ADDR,~ByteToSegs(Counter));

      ++Counter;

   }


}
