/* Embedded PC'S ISA Bus Chapter 7 -- Ed Nisley                              */
/* Micro-C program to exercise Firmware Development Board battery-backed RAM */
/* You =must= remove the system Reset connector and -IOCHCHK jumper          */
/*  these tests do not twiddle the watchdog timer output bit                 */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

#define MAX_ERRORS   10          /* stop testing after this many mismatches  */

/*---------------------------------------------------------------------------*/
/* The main event                                                            */

char Option;
WORD Counter;
WORD RAMSeg;
WORD RAMAddr;
WORD RAMSize;
BYTE TestData;
int  ErrorCounter;

BYTE PRSData;
BYTE *pPRSData;
WORD PRSSegment;

BYTE Alarm;

main() {

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("\nBattery-backed RAM Tests\n");
   putstr("Embedded PC's ISA Bus Chapter 7 -- Ed Nisley\n");

   putstr("Calculating pseudorandom data...  ");
asm {
   MOV   >RANDSEED,#$0001
}
   PRSSegment = GetDataSeg() + 0x1000;             /* step 64K above data    */
   for (pPRSData = 0; pPRSData != 0x8000; pPRSData += 2) {
      pokew(PRSSegment,pPRSData,rand(0x8000));
   }
   putstr("done\n");

   while (TRUE) {

      putstr("\n\nMemory checkout routines:\n");
      putstr("== Remove System Reset connection and -IOCHCHK jumper! ==\n");
      putstr(" 1 - Wiggle watchdog & RAM controls\n");
      putstr(" 2 - Repeated memory reads\n");
      putstr(" 3 - Repeated memory writes (=NOT= for EEPROM!)\n");
      putstr(" 4 - Load 32K RAM w/PRS32K data (check w/#5)\n");
      putstr(" 5 - Verify 32K PRS32K data\n");
      putstr("Select option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '1' :
         putstr("Observe watchdog & RAM write enable LEDs\n");
         putstr(" ... watchdog Reset output should be inactive\n");
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            outpw(CTLS_ADDR,Counter);
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(HI(Counter)));       /* use high byte */
            if (BIOSWait(0x0000,0x01F4)) {         /* delay for 500 us       */
               putstr("BIOS delay is broken!\n");
               break;
            }
            ++Counter;
         }
         outpw(CTLS_ADDR,0x0000);                  /* reset to default       */
         break;
      case '2' :
         RAMSize = NV_SIZE;                        /* set up for assembler   */
         RAMSeg = NV_SEGMENT;
         printf("Reading memory at segment %04x, size %04x\n",
                RAMSeg,RAMSize);
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
asm {
   MOV   CX,RAMSize        repeat for whole block
   PUSH  DS
   MOV   AX,RAMSeg         set up RAM address
   MOV   DS,AX
   XOR   SI,SI
   REP
   LODSB
   POP   DS
}
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(Counter));
            ++Counter;
         }
         break;
      case '3' :
         printf("Writing memory at segment %04x, size %04x\n",
                NV_SEGMENT,NV_SIZE);
         RAMAddr = 0x0000;
         Counter = 0;
         outpw(CTLS_ADDR,NV_WENABLE);              /* allow writes           */
         while (!chkch()) {
            if (RAMAddr == NV_SIZE) {
               outp(SYNC_ADDR,0x01);               /* sync on printer port   */
               outpw(LED_ADDR,~ByteToSegs(Counter));
               ++Counter;
               RAMAddr = 0;
            }
            outp(SYNC_ADDR,0x02);                  /* mark each peek         */
            poke(NV_SEGMENT,RAMAddr,~RAMAddr);     /* write inverse address  */
            outp(SYNC_ADDR,0x00);
            ++RAMAddr;
         }
         outpw(CTLS_ADDR,0x0000);                  /* reset port             */
         break;
      case '4' :
         printf("Loading 32K RAM at seg %04x\n",NV_SEGMENT);
         outpw(CTLS_ADDR,NV_WENABLE);              /* allow writes           */
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != NV_SIZE; ++RAMAddr) {
            PRSData = peek(PRSSegment,RAMAddr);
            outp(SYNC_ADDR,0x01);                  /* mark the write         */
            poke(NV_SEGMENT,RAMAddr,PRSData);      /* do it                  */
            TestData = peek(NV_SEGMENT,RAMAddr);
            if (TestData != PRSData) {
               printf("Cannot set %04x to %02x, is %02x\n",
                  RAMAddr,PRSData,TestData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(HI(RAMAddr)));
         }
         outpw(CTLS_ADDR,0x0000);                  /* reset port             */
         putstr("... done\n");
         break;
      case '5' :
         printf("Verifying 32K block at seg %04x\n",NV_SEGMENT);
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != NV_SIZE; ++RAMAddr) {
            outp(SYNC_ADDR,0x01);                  /* mark this loop         */
            TestData = peek(NV_SEGMENT,RAMAddr);
            PRSData = peek(PRSSegment,RAMAddr);
            if (PRSData != TestData) {
               printf("Mismatch at %04x, data is %02x, should be %02x\n",
                       RAMAddr,TestData,PRSData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
         }
         putstr("... done\n");
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
