/* Embedded PC's ISA Bus Chapter 8 -- Ed Nisley                              */
/* Micro-C program to exercise Firmware Development Board BIOS extensions    */
/* You must do these steps in this order...                                  */
/*  - create a boot diskette with the appropriate size of LOADEXT.SEC        */
/*     ... a 720K diskette uses LDEX720.SEC, for example                     */
/*  - copy FDBEXT.BIN onto the diskette                                      */
/*  - hold the button on the Firmware Development Board & boot the diskette  */
/*     ... this loads the extension and calculates the checksum              */
/*  - create a boot diskette with the appropriate size of BOOTSECT.SEC       */
/*     ... a 720K diskette uses BOOT720.SEC, for example                     */
/*  - copy this program (EXTTEST.BIN) to that diskette as the first file     */
/*  - put the diskette in the drive and boot with the pushbutton released    */
/*     ... this program will start running                                   */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

/*---------------------------------------------------------------------------*/
/* Layout of the variables and suchlike in the BIOS extension's segment      */
/* We use the adjusted segment value so our offsets match up                 */
/*  C800:0400 = C7F0:0500 which matches the default COM file layout          */

#define EXTSEG 0xC7F0                  /* first byte in RAM is at 0100...    */
#define VAROFF 0x0500

#define SIGNATURE_LSW   (VAROFF+0x0000)
#define SIGNATURE_MSW   (VAROFF+0x0002)

#define STARTUP_OFF     (VAROFF+0x0004)
#define STARTUP_SEG     (VAROFF+0x0006)

#define RESET_CTR       (VAROFF+0x0008)
#define STARTUP_CTR     (VAROFF+0x000A)
#define NMI_CTR         (VAROFF+0x000C)

#define TICK_CTR_LSW    (VAROFF+0x000E)
#define TICK_CTR_MSW    (VAROFF+0x0010)

#define OLD_NMI         (VAROFF+0x0012)
#define OLD_TIMER       (VAROFF+0x0016)

#define INIT_CS         (VAROFF+0x001A)


/*---------------------------------------------------------------------------*/
/* The main event                                                            */

main() {

char  Option;
WORD  LastTick;
BYTE  SysCtls;

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("BIOS Extension exerciser\n");
   putstr("Embedded PC's ISA Bus Chapter 8 -- Ed Nisley\n");

   printf("Values in non-volatile memory at %04x:%04x\n",
         EXTSEG,VAROFF);
   printf(" Signature         %04x%04x\n",
         peekw(EXTSEG,SIGNATURE_MSW),peekw(EXTSEG,SIGNATURE_LSW));
   printf(" Reset counter     %u\n",peekw(EXTSEG,RESET_CTR));
   printf(" Startup counter   %u\n",peekw(EXTSEG,STARTUP_CTR));
   printf(" NMI counter       %u\n",peekw(EXTSEG,NMI_CTR));
   printf(" Initial CS        %04x\n",peekw(EXTSEG,INIT_CS));
   putch('\n');

   printf("NMI Vector %04x:%04x\n",
          peekw(0x0000,0x000A),peekw(0x0000,0x0008));
   SysCtls = inp(SYS_CTLS);
   printf(" SYS_CTLS port is %02x\n",SysCtls);
   if (SysCtls & 0x08) {
      printf(" Enabling NMI...");
      outp(SYS_CTLS,SysCtls & ~0x08);
      printf(" done\n\n");
   }

   printf("Changing startup vector from %04x:%04x to %04x:%04x...\n",
         peekw(0x0040,0x0069),peekw(0x0040,0x0067),
         peekw(EXTSEG,STARTUP_SEG),peekw(EXTSEG,STARTUP_OFF));
   pokew(0x0040,0x0067,peekw(EXTSEG,STARTUP_OFF));
   pokew(0x0040,0x0069,peekw(EXTSEG,STARTUP_SEG));
   putstr(" setting shutdown code...\n");
   disable();
   outp(RTC_ADDR,0x0f);                /* aim at shutdown reason code        */
   outp(RTC_DATA,0x0A);                /*  ... vector through 0040:0067      */
   enable();
   putstr(" done\n");

   putstr("\nAdjust PFI pot to trigger NMI, check LED display\n");
   putstr("Press 1 to trigger soft shutdown and restart\n");
   putstr("If timer tick counter does not change, are the extensions ON?\n");

   LastTick = 0xFFFF;

   while (TRUE) {

      if (LastTick != peekw(EXTSEG,TICK_CTR_LSW)) {
         LastTick = peekw(EXTSEG,TICK_CTR_LSW);
         printf("\r%04x%04x  ",
                peekw(EXTSEG,TICK_CTR_MSW),peekw(EXTSEG,TICK_CTR_LSW));
      }

      Option = chkch();

      switch (Option) {
      case '1' :
         outp(KEY_CMD,0xFE);           /* trigger soft shutdown and restart  */
         break;
      case 0 :
         break;
      default :
         putstr("Invalid option...\n");
      }

   }

}
