; BIOS Extension for Firmware Development Board
; Must be loaded into EEPROM/RAM with the LOADEXT.ASM boot loader
; Illustrates interrupt hooks and other useful tricks

; *** NOTE ***
; This code uses 386/486-specific opcodes
; It will =not= run on 8088 or 80286 machines!

; Embedded PC's ISA Bus Chapter 8 -- Ed Nisley
; Assemble with Borland TASM

              .386C                         ; allow 386 opcodes

              IDEAL
              MODEL     TINY

;--------------
; Set up a few constants

COUNT_NMI     EQU       0                   ; nonzero to count powerdown delay
USE_IOCHCK    EQU       0                   ; nonzero to test IOCHCK for NMI

LED_ADDR      EQU       031Eh               ; LED digits
SW_ADDR       EQU       031Eh               ; DIP switch inputs

CTLS_ADDR     EQU       031Ch               ; control output port address
WATCHDOG      EQU       0100h               ; watchdog toggle bit
NVR_WENABLE   EQU       0200h               ; write-enable bit

STAT_ADDR     EQU       031Ch               ; status input port address
PWR_GOOD      EQU       0100h               ; 1 = power still OK
PUSHBUTTON    EQU       0200h               ; pushbutton input bit 0 = pushed

RTC_ADDR      EQU       0070h               ; real-time clock address port
RTC_DATA      EQU       0071h               ;  ... and data

KEY_DATA      EQU       60h                 ; keyboard controller data/status

SYS_CTLS      EQU       61h                 ; system control/status port
IOCHCK        EQU       40h                 ;  1 when IOCHCK is active

KEY_CMD       EQU       64h                 ; keyboard controller commands


;--------------
; Handy macros and suchlike

              UNION     FARPTR

DWORD         DD        ?                   ; both words at once

              STRUC                         ; seg & offset separately
OFFP          DW        ?
SEGP          DW        ?
              ENDS

              ENDS

;--------------
; Variables live in the next block of our RAM
; They must be defined before they're used, but they must live after the code
;  ... so we use an absolute segment to lay 'em out
; We =must= enable and disable RAM writes around each update!

              SEGMENT   VarBlock AT 0C7Fh

              ORG       0500h               ; skip over the code block

Signature     DD        ?                   ; should be BABEFACE in binary

StartupVector FARPTR    {?}                 ; vector to StartupHandler

ResetCtr      DW        ?                   ; number of resets (never zeroed)
StartupCtr    DW        ?                   ; number of soft resets
NMICtr        DW        ?                   ; number of NMIs

TickCtr       DD        ?                   ; number of BIOS timer ticks

OldTimer      FARPTR    {?}                 ; old timer tick vector
OldNMI        FARPTR    {?}                 ; old NMI handler

InitCS        DW        ?                   ; CS on entry

              ENDS

;-------------
; This does not start at offset 0100, so be careful about data accesses!
; Actual execution is at absolute address C800:0003...
; The first three bytes are the signature and length

              CODESEG
              STARTUPCODE

              DB        055h                ; signature
              DB        0AAh
              DB        2                   ; length in units of 512 bytes

              JMP       SHORT BootEntry     ; force two-byte jump

              DB        00h                 ; loader sets this checksum byte

;--- constants that must be stored within the checksummed region

RevCode       DW        0001h               ; current revision level

;--- if pushbutton is down, exit without doing much

BootEntry:
              MOV       DX,STAT_ADDR
              IN        AX,DX
              TEST      AX,PUSHBUTTON
              JNZ       SHORT Continue      ; nonzero means not pushed

              MOV       AX,NOT 0101h        ; show -- to track our path
              MOV       DX,LED_ADDR
              OUT       DX,AX

              RETF                          ; return to normal BIOS boot

Continue:

;--- adjust CS and DS to simulate the normal COM situation
;    we need both code and data starting at offset 0100 rather than 0000
;    C800:0000 is also C7F0:0100, so we just subtract 0010 from the segments
;    Storing this CS in the vectors allows normal access after an interrupt

              MOV       AX,NOT 0100h        ; show single - here
              MOV       DX,LED_ADDR
              OUT       DX,AX

              MOV       SI,CS               ; save this for later

              MOV       AX,CS
              SUB       AX,0010h            ; adds 100 to offsets in segment
              PUSH      AX
              PUSH      OFFSET BootStart
              RETF                          ; set CS and IP to new values

BootStart:
              MOV       AX,0057Eh           ; show r0 on LEDs to mark entry
              CALL      ShowBits

              PUSH      DS                  ; preserve seg regs
              PUSH      ES
              PUSH      FS
              PUSH      GS

              MOV       AX,CS               ; set up DS to match CS
              MOV       DS,AX
              MOV       ES,AX

              ASSUME    ES:VarBlock

              MOV       AX,0                ; FS points to 0000:xxxx
              MOV       FS,AX

              MOV       AX,0040h            ; GS points to 0040:xxxx
              MOV       GS,AX

              CALL      OpenRAM             ; enable writes

              MOV       [ES:InitCS],SI      ; save this for ExtTest
              INC       [ES:ResetCtr]       ; count this reset

;--- capture BIOS timer interrupt to count elapsed time

              MOV       AX,00530h           ; r1
              CALL      ShowBits

              MOV       [ES:TickCtr],0

              MOV       BX,01Ch             ; capture BIOS end-of-tick handler
              MOV       [ES:OldTimer.OFFP],OFFSET TickHandler
              MOV       [ES:OldTimer.SEGP],CS
              MOV       DI,OFFSET OldTimer
              CALL      SwapVector

;--- hook NMI vector and enable NMI
;    the BIOS may disable NMI during the rest of the boot sequence!

              MOV       AX,0576Dh            ; r2
              CALL      ShowBits

              MOV       BX,02h              ; capture NMI vector
              MOV       [ES:OldNMI.OFFP],OFFSET NMIHandler
              MOV       [ES:OldNMI.SEGP],CS
              MOV       DI,OFFSET OldNMI
              CALL      SwapVector

              IN        AX,SYS_CTLS
              AND       AL,37h              ; bit 3 = 0 to enable NMI
              OUT       SYS_CTLS,AL


;--- put our startup handler address in RAM for the test routine

              MOV       [ES:StartupVector.OFFP],OFFSET StartupHandler
              MOV       [ES:StartupVector.SEGP],CS


;--- write-protect the RAM and return to the BIOS

              CALL      CloseRAM

              POP       GS                  ; restore seg regs
              POP       FS
              POP       ES
              POP       DS

              MOV       AX,00505h           ; show rr to indicate the end
              CALL      ShowBits

              RETF                          ; and return to BIOS!

;-------------
; Enable RAM writes
; Clears the watchdog bit in passing

              PROC      OpenRAM

              MOV       DX,CTLS_ADDR
              MOV       AX,NVR_WENABLE
              OUT       DX,AX

              RET

              ENDP      OpenRAM


;-------------
; Disable RAM writes
; Sets the watchdog bit in passing

              PROC      CloseRAM

              MOV       DX,CTLS_ADDR
              MOV       AX,WATCHDOG
              OUT       DX,AX

              RET

              ENDP      CloseRAM


;-------------
; Write AX to the LED digits

              PROC      ShowBits

              PUSH      DX

              MOV       DX,LED_ADDR
              NOT       AX                  ; make 1 bits light LEDs
              OUT       DX,AX

              POP       DX

              RET

              ENDP      ShowBits


;--------------
; Timer tick handler
; Counts the interrupt and vectors to old handler
; We use CS because we set up the vector so this works...

              PROC      TickHandler

              PUSH      AX
              PUSH      DX

              CALL      OpenRAM

              INC       [CS:TickCtr]

              CALL      CloseRAM

              POP       DX
              POP       AX

              JMP       SMALL [CS:OldTimer.DWORD]

              ENDP      TickHandler

;--------------
; NMI handler
; This ensures a clean shutdown before a power failure
; PWR_GOOD may be high even after -IOCHCK triggers an interrupt, so...
;  we have an option to check for just the IOCHCK bit

              PROC      NMIHandler

              PUSH      AX
              PUSH      DX

              CALL      OpenRAM

              INC       [CS:NMICtr]

              CALL      CloseRAM

             IF         COUNT_NMI

              MOV       CX,0                ; set up the counter

@@Wait:
              MOV       DX,LED_ADDR         ; show the loop counter in binary
              MOV       AX,CX
              NOT       AX
              OUT       DX,AX

             IF         USE_IOCHCK

              IN        AL,SYS_CTLS         ; look at IOCHCK flag
              TEST      AL,IOCHCK
@@Kaput:      JNZ       SHORT @@Kaput       ; lock up when it goes high

             ELSE

              MOV       DX,STAT_ADDR        ; check power status
              IN        AX,DX
              TEST      AX,PWR_GOOD
@@Kaput:      JZ        SHORT @@Kaput       ; lock up when status goes low

             ENDIF

              INC       CX
              JMP       @@Wait


             ELSE

              MOV       AX,1510h            ; show ni (more or less)
              MOV       DX,LED_ADDR
              NOT       AX
              OUT       DX,AX

             IF         USE_IOCHCK

              IN        AL,SYS_CTLS         ; look at IOCHCK flag
              TEST      AL,IOCHCK
              JNZ       SHORT @@Lockup      ; lock up when it goes high

             ELSE

              MOV       DX,STAT_ADDR        ; check power status
              IN        AX,DX
              TEST      AX,PWR_GOOD
              JZ        SHORT @@Lockup      ; zero = power NOT good...

             ENDIF

             ENDIF

              POP       DX
              POP       AX

              JMP       SMALL [CS:OldNMI.DWORD]

@@Lockup:
              MOV       AX,08080h           ; both decimal points
              CALL      ShowBits

@@Stall:      JMP       @@Stall

              ENDP      NMIHandler


;--------------
; Startup handler
; This is =not= an interrupt handler!
; It is called by the BIOS just after a soft reset
; We tick a counter and do another, real reset to start over again

              PROC      StartupHandler

              CALL      OpenRAM

              INC       [CS:StartupCtr]     ; record another startup

              CALL      CloseRAM

              MOV       AX,00580h           ; show r.
              CALL      ShowBits

              MOV       AL,0FEh             ; tell kbd controller
              OUT       KEY_CMD,AL          ;  ... to blip the reset lind
@@Stall:      JMP       @@Stall

              ENDP      StartupHandler

;--------------
; Swap old and new vectors
; BL has the vector number
; [DS:DI] points to the new vector and gets the old one
; Assumes FS is zero...

              PROC      SwapVector

              MOV       BH,0                ; ensure high byte is zero
              SHL       BX,2                ; convert Int ID to offset

              PUSHF                         ; no interrupts just now
              CLI

              XCHG      AX,[DI]             ; get new vector
              XCHG      DX,[DI+2]

              XCHG      AX,[FS:BX]          ; swap with old vector
              XCHG      DX,[FS:BX+2]

              XCHG      AX,[DI]             ; save old vector
              XCHG      DX,[DI+2]

              POPF                          ; restore interrupts

              RET

              ENDP      SwapVector

;--------------
; Reserve a block for the variables
; This is =not= included in the BIOS extension block size
;  ... thus, it does =not= count in the BIOS checksum calculation
; The extension loader writes the RAM based on the file size
;  ... thus, this =does= count in the data that's transferred to RAM

              ORG       0500h               ; *** must be after code block

              DD        0BABEFACEh          ; signature to mark good load

              DW        252 DUP (0)         ; zero everything else

              END

