/* Embedded PC's ISA Bus Chapter 9 -- Ed Nisley                              */
/* Micro-C program to exercise character LCD                                 */
/* You =must= remove the system Reset connector and -IOCHCHK jumper          */
/*  these tests do not twiddle the watchdog timer output bit                 */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

#define TRC_SCROLLxxx
#define TRC_BSxxx

/*---------------------------------------------------------------------------*/
/* LCD constants                                                             */
/* These assume a standard Hitachi 44780 LCD controller chip                 */

/* Inline loop delay counter between E output and data readback              */
/*   300 33 MHz '386SX                                                       */
/*  1500 80 MHz '486DX2                                                      */
#define RD_DELAY    1500         /* tune for reliable vertical scrolling!    */

#define VISIBLEROWS 4            /* number of visible rows                   */
#define VISIBLECOLS 20           /* number of visible columns                */

#define LCDCURSOFF   0                             /* for LCDSelCursor       */
#define LCDCURSON    1

#define LCDDISPOFF   0                             /* for LCDSelDisplay      */
#define LCDDISPON    1
                                                   /* for LCDSendByte        */
#define SENDCMD      0x00                          /* bit 0 sets cmd/data    */
#define SENDDATA     0x01
#define SENDFORCE    0x02                          /* bit 1 sets wait/force  */
#define SENDWAIT     0x00                          /* no bit set!            */

#define LCD_VSCROLL  0x01                          /* scroll after last line */
#define LCD_BLANKING 0x02                          /* blank during scrolls   */
#define LCD_CRLF     0x04                          /* force CR after \n      */
#define LCD_HWRAP    0x08                          /* wrap at end of line    */
#define LCD_DESTBS   0x10                          /* erase char on BS       */

#define CONTROLDEFS  LCD_DESTBS + LCD_CRLF + LCD_VSCROLL + LCD_BLANKING;
#define CURSORDEF    (0x0C + 1)  /* + 1=blink block, 2=underline, 3=both     */

/*---------------------------------------------------------------------------*/
/* LCD variables                                                             */

int RowStarts[4];                      /* starting address of each row       */
BYTE RowBuffer[81];                    /* holds row of data + term byte      */

int LCDControls;

int CurrentRow;
int CurrentCol;
int CursorState;


/*---------------------------------------------------------------------------*/
/* Macro to simplify life                                                    */

#define LCDSendCmd(x) LCDSendByte(x,SENDWAIT|SENDCMD)

/*---------------------------------------------------------------------------*/
/* Read data byte from controller                                            */
/* Remember that the LCD cursor position advances when you read data         */

int LCDReadByte(Mode)
WORD Mode;
{
WORD Data;
WORD PortValue;

   outp(SYNC_ADDR,inp(SYNC_ADDR) | 0x02);          /* scope sync             */

   PortValue = (SENDDATA == Mode) ? LCD_RS : 0;    /* read data or addr?     */
   PortValue |= LCD_RW + LCD_NODRV;    /* LCD read and disable drivers       */
   outpw(CTLS_ADDR,PortValue);

   PortValue |= LCD_E;                 /* strobe data from LCD               */
   outpw(CTLS_ADDR,PortValue);

asm {
   MOV   CX,#RD_DELAY                  ; delay to allow data settling time
?lp1 LOOP ?lp1
}

   Data = inpw(STAT_ADDR) & 0x00FF;    /* fetch and isolate LCD data         */

   PortValue &= ~LCD_E;                /* remove LCD strobe                  */
   outpw(CTLS_ADDR,PortValue);

   outp(SYNC_ADDR,inp(SYNC_ADDR) & ~0x02);         /* sync off               */

   return Data;

}



/*---------------------------------------------------------------------------*/
/* Toggle Enable line to strobe data or commands into the LCD controller     */
/* PortValue gives the current state of the output bits                      */

int BlipEnable(PortValue)
WORD PortValue;
{

   PortValue |= LCD_E;                             /* set high               */
   outpw(CTLS_ADDR,PortValue);

   PortValue &= ~LCD_E;                            /* and low again          */
   outpw(CTLS_ADDR,PortValue);

   return PortValue;

}

/*---------------------------------------------------------------------------*/
/* Wait for Busy flag to clear                                               */
/* Returns only the address part of the byte, as busy is always off...       */
/* This will hang if the LCD is disconnected and the bit floats high!        */

int LCDWaitBusy() {
WORD Data;

   Data = 0;
   do {
      outp(SYNC_ADDR,inp(SYNC_ADDR) | 0x04);       /* scope sync             */
      Data = LCDReadByte(SENDCMD);
      outp(SYNC_ADDR,inp(SYNC_ADDR) & ~0x04);
   } while (Data & 0x0080);

   return Data;

}



/*---------------------------------------------------------------------------*/
/* Send command or data byte to the LCD controller                           */
/* If this is data, the LCD cursor position will advance                     */

LCDSendByte(Data,Mode)
WORD Data;
WORD Mode;
{
WORD PortValue;

   if (!(Mode & SENDFORCE)) {          /* want to force without wait?        */
      LCDWaitBusy();
   }

   PortValue = (Mode & SENDDATA) ? LCD_RS : 0;     /* cmd/data, -WR, enabled */
   PortValue |= Data;                              /* combine data           */
   outpw(CTLS_ADDR,PortValue);

   BlipEnable(PortValue);

   PortValue |= LCD_RW;                            /* disable -Wr line       */
   outpw(CTLS_ADDR,PortValue);

}




/*---------------------------------------------------------------------------*/
/* Initialize LCD hardware and set display control variables                 */

LCDInitialize() {

WORD PortValue;

   outp(SYNC_ADDR,inp(SYNC_ADDR) | 0x08);          /* scope sync             */

   CurrentRow = CurrentCol = 0;
   LCDControls = CONTROLDEFS;          /* set initial control state          */
   RowStarts[0] = 0x00;
   RowStarts[1] = 0x40;
   RowStarts[2] = 0x14;
   RowStarts[3] = 0x54;

   outpw(CTLS_ADDR,0);                 /* clear all lines                    */

/*--- first few commands cannot check BUSY, as state is not known            */

   BIOSWait(0x0000,0x3A98);            /* delay for 15 ms (after power on!)  */
   PortValue = 0x0030;                 /* 8-bit mode, ctls low, enabled      */
   outpw(CTLS_ADDR,PortValue);
   BlipEnable(PortValue);

   BIOSWait(0x0000,0x2710);            /* delay for >4.1 ms (use 10 ms)      */
   BlipEnable(PortValue);              /*  ... send same command again       */

   BIOSWait(0x0000,0x03E8);            /* delay for >100 us (use 1 ms)       */
   BlipEnable(PortValue);              /*  ... send same command again       */

   BIOSWait(0x0000,0x03E8);            /* unspecified delay (use 1 ms)       */

   PortValue = 0x0038;                 /* 8-bit mode, 2 lines, 5x7 font      */
   BlipEnable(PortValue);

/*--- OK to check BUSY for the remaining setup commands                      */

   LCDSendCmd(0x08);                               /* disp/cur/blink off     */
   LCDSendCmd(0x01);                               /* clear and home         */
   LCDSendCmd(0x06);                               /* increment, no shift    */

   CursorState = CURSORDEF;                        /* set up cursor          */
   LCDSendCmd(CursorState);                        /* ... with display ON    */

   outp(SYNC_ADDR,inp(SYNC_ADDR) & ~0x08);

}


/*---------------------------------------------------------------------------*/
/* Set LCD cursor to row/column location                                     */
/* Rows and columns start from 0 in the upper-left corner                    */

LCDSetCursor(Row,Col)
int Row;
int Col;
{
WORD CursorCmd;

   Row = max(Row,0);                   /* prevent wrap off top               */
   Col = max(Col,0);

   CurrentRow = Row = min(Row,VISIBLEROWS-1);      /* and off bottom         */
   CurrentCol = Col = min(Col,VISIBLECOLS-1);

   CursorCmd = RowStarts[Row] + Col | 0x80;        /* 0x80 sets cursor       */

   LCDSendCmd(CursorCmd);

}


/*---------------------------------------------------------------------------*/
/* Set cursor to 0,0 but do not clear the display                            */

LCDHome() {

   CurrentRow = CurrentCol = 0;

   LCDSendCmd(0x02);

}


/*---------------------------------------------------------------------------*/
/* Clear display and home the cursor                                         */

LCDClear() {

   CurrentRow = CurrentCol = 0;

   LCDSendCmd(0x01);

}


/*---------------------------------------------------------------------------*/
/* Scroll display up N lines                                                 */

LCDScrollUp(NumLines)
WORD NumLines;
{
WORD SaveRow;
WORD SaveCol;
WORD ThisRow;
WORD PrevRow;
WORD ThisCol;

   if (!NumLines) {                    /* this makes no sense...             */
      return;
   }

   if (NumLines >= VISIBLEROWS) {      /* this is easy                       */
      LCDClear();
      return;
   }

   if (LCDControls & LCD_BLANKING) {   /* blank during scrolling?            */
      LCDSelDisplay(LCDDISPOFF);
   }

   LCDWaitBusy();
   SaveRow = CurrentRow;
   SaveCol = CurrentCol;

   outp(SYNC_ADDR,inp(SYNC_ADDR) | 0x10);          /* scope sync             */

#ifdef TRC_SCROLL
   putstr("Scroll readback:\n");
#endif

   for (ThisRow = NumLines, PrevRow = 0;           /* from and to targets    */
        ThisRow < VISIBLEROWS;
        ++ThisRow, ++PrevRow) {
      LCDSetCursor(ThisRow,0);                     /* fetch whole row        */
      for (ThisCol = 0; ThisCol < VISIBLECOLS; ++ThisCol) {
         RowBuffer[ThisCol] = LCDReadByte(SENDDATA);
      }
#ifdef TRC_SCROLL
      RowBuffer[ThisCol] = 0;                      /* add string delimiter   */
      printf(" Row %1u: %s\n",ThisRow,RowBuffer);    /* show results           */
#endif
      LCDSetCursor(PrevRow,0);                     /* move to target         */
      for (ThisCol = 0; ThisCol < VISIBLECOLS; ++ThisCol) {
         LCDSendByte(RowBuffer[ThisCol],SENDWAIT|SENDDATA);
      }
   }



   outp(SYNC_ADDR,inp(SYNC_ADDR) & ~0x10);

   LCDSetCursor(VISIBLEROWS-1,0);                  /* set last row to blank  */
   for (ThisCol = 0; ThisCol < VISIBLECOLS; ++ThisCol) {
      LCDSendByte(' ',SENDWAIT|SENDDATA);
   }

   LCDSetCursor(SaveRow,SaveCol);                  /* restore cursor loc     */
   LCDSelDisplay(LCDDISPON);                       /* always enable display  */

}


/*---------------------------------------------------------------------------*/
/* Enable/disable LCD cursor                                                 */
/* Side effect is to turn the display on whenever this happens               */

LCDCtlCursor(State)
WORD State;
{
   switch (State) {
   case LCDCURSOFF :
      CursorState = 0x0C;              /* no cursor display at all           */
      break;
   case LCDCURSON :
      CursorState = CURSORDEF;         /* use the default                    */
      break;
   }

   LCDSendCmd(CursorState);


}



/*---------------------------------------------------------------------------*/
/* Enable/disable LCD panel                                                  */

LCDSelDisplay(State)
WORD State;
{
WORD DispCmd;

   DispCmd = CursorState & ~0x04;                  /* force display OFF      */

   DispCmd |= (LCDDISPON == State) ? 0x04 : 0;     /* set ON or OFF          */

   LCDSendCmd(DispCmd);


}



/*---------------------------------------------------------------------------*/
/* Send printable chars to the LCD                                           */
/* Handles standard control chars                                            */

LCDSendChar(Char)
char Char;
{
WORD ThisCol;
WORD ThisChar;

   switch (Char) {
   case LF :
      if (CurrentRow == (VISIBLEROWS - 1)) {       /* on last row?           */
         if (LCDControls & LCD_VSCROLL) {          /* yes, want scrolling?   */
            LCDScrollUp(1);                        /* yes, do it & stay put  */
         }
         else {                                    /* no scrolling           */
            CurrentRow = 0;                        /*  so return to top      */
         }
      }
      else {                                       /* not on last row        */
         ++CurrentRow;                             /* so just step down one  */
      }
      if (!(LCDControls & LCD_CRLF)) {             /* want CR after newline? */
         LCDSetCursor(CurrentRow,CurrentCol);      /*  ... no, so done       */
         break;
      }
      /* else fall through */
   case CR :
      LCDSetCursor(CurrentRow,0);                  /* to leftmost column     */
      break;
   case FORMFEED :
      LCDClear();
      break;
   case BS :
      if (CurrentCol) {                            /* if not at edge         */
         --CurrentCol;                             /*   back up one col      */
      }
      LCDSetCursor(CurrentRow,CurrentCol);         /* adjust actual cursor   */
      if (!(LCDControls & LCD_DESTBS)) {           /* want char destruction? */
         break;
      }
      /* else fall through */
   case DEL :
      LCDWaitBusy();
      if (LCDControls & LCD_BLANKING) {            /* blank during scroll?   */
         LCDSelDisplay(LCDDISPOFF);
      }

      for (ThisCol = CurrentCol; ThisCol < VISIBLECOLS; ++ThisCol) {
         RowBuffer[ThisCol] = LCDReadByte(SENDDATA);
      }

#ifdef TRC_BS
      RowBuffer[ThisCol] = 0;
      printf("BS readback at %2u: [%s]\n",CurrentCol,RowBuffer+CurrentCol);
#endif

      LCDSetCursor(CurrentRow,CurrentCol);         /* restore cursor loc     */

      for (ThisCol = CurrentCol+1; ThisCol < VISIBLECOLS; ++ThisCol) {
         ThisChar = RowBuffer[ThisCol];            /* slide over 1           */
         LCDSendByte(ThisChar,SENDWAIT|SENDDATA);
      }
      LCDSendByte(' ',SENDWAIT|SENDDATA);
      LCDSetCursor(CurrentRow,CurrentCol);         /* restore cursor loc     */
      LCDSelDisplay(LCDDISPON);                    /* always enable display  */
      break;
   default :
      LCDSendByte(Char,SENDWAIT|SENDDATA);         /* send the char          */
      ++CurrentCol;                                /* tick column counter    */
      if (CurrentCol >= VISIBLECOLS) {             /* past end of line?      */
         LCDSetCursor(CurrentRow,0);               /* reset to left edge     */
         if (!(LCDControls & LCD_HWRAP)) {         /* go to next line?       */
            LCDSendChar(LF);                       /* yes, so force it       */
         }
      }
   }

}




/*---------------------------------------------------------------------------*/
/* Send C string to the LCD                                                  */

LCDSendString(pString)
char *pString;
{

   while (0 != *pString) {
      LCDSendChar(*pString);
      ++pString;
   }

}

/*---------------------------------------------------------------------------*/
/* Show prompt on serial output & wait for key                               */

ShowPrompt(pString)
char *pString;
{

   putstr(pString);
   getch();
   putch('\n');
}




/*---------------------------------------------------------------------------*/
/* The main event                                                            */

char Option;
WORD Counter;
int SerChar;

main() {

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("\nCharacter LCD panel tests\n");
   putstr("Embedded PC's ISA Bus Chapter 9 -- Ed Nisley\n\n");

   printf("LCD should have %u rows and %u columns visible\n",
            VISIBLEROWS,VISIBLECOLS);

   while (TRUE) {

      putstr("\n\nHardware checkout routines:\n");
      putstr("== Remove System Reset connection and -IOCHCHK jumper! ==\n");
      putstr(" 1 - Initialize LCD in loop\n");
      putstr(" 2 - Write counting sequence\n");
      putstr(" 3 - Verify cursor and LCD controls\n");
      putstr(" 4 - Echo serial characters to LCD\n");
      putstr("Select option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '1' :
         putstr("Initialization loop...\n");
         Counter = 0;
         while (!chkch()) {
            LCDInitialize();
            if (BIOSWait(0x0003,0xD090)) {         /* delay for 250 ms       */
               putstr("BIOS delay is broken!\n");
               break;
            }
            outpw(LED_ADDR,~ByteToSegs(Counter));  /* show counter           */
            ++Counter;
         }
         break;
      case '2' :
         putstr("Ascending chars with vertical scrolling\n");
         LCDInitialize();
         Counter = ' ';
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            LCDSendChar(Counter);
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(Counter));
            if (BIOSWait(0x0003,0xD090)) {         /* delay for 250 ms       */
               putstr("BIOS delay is broken!\n");
               break;
            }
            Counter = (Counter >= 0x7F) ? ' ' : Counter + 1;
         }
         break;
      case '3' :
         putstr("Cursor positioning...\n");
         LCDInitialize();
         switch (VISIBLEROWS) {
         case 4 :
            LCDSetCursor(3,5);
            LCDSendString("3,5");
         case 3 :
            LCDSetCursor(2,10);
            LCDSendString("2,10");
         case 2 :
            if (VISIBLECOLS > 16) {
               LCDSetCursor(1,16);
               LCDSendString("1,16");
            }
            else {
               LCDSetCursor(1,3);
               LCDSendString("1,3");
            }
         case 1 :
            LCDSetCursor(0,0);
            LCDSendString("0,0");
            LCDSetCursor(0,8);
         }
         LCDSendString("See CRT!");
         ShowPrompt("Positions OK? Press Enter");
         LCDSelDisplay(LCDDISPOFF);
         ShowPrompt("LCD should be off; press Enter");
         LCDSelDisplay(LCDDISPON);
         LCDCtlCursor(LCDCURSOFF);
         ShowPrompt("LCD cursor should be off; press Enter");
         LCDCtlCursor(LCDCURSON);
         LCDHome();
         ShowPrompt("LCD cursor should be at (0,0); press Enter");
         LCDClear();
         ShowPrompt("LCD should be cleared; press Enter");
         LCDSendString("Backspace chars");
         for (Counter=0; Counter < 15; ++Counter) {
            LCDSendChar('\x08');
            BIOSWait(0x0007,0xA120);
         }
         BIOSWait(0x000F,0x4240);
         LCDSendString("Delete char\r");
         BIOSWait(0x000F,0x4240);
         for (Counter=0; Counter < 11; ++Counter) {
            LCDSendChar('\x7F');
            BIOSWait(0x0007,0xA120);
         }
         LCDSendString("        CR char\rDone!");
         putstr("Done!\n");
         break;
      case '4' :
         putstr("Press ESC to bail out...\n");
         LCDInitialize();
         do {
            SerChar = getch();
            LCDSendChar(SerChar);
            putch(SerChar);
            outpw(LED_ADDR,~ByteToSegs(SerChar));
         } while (SerChar != ESC);
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
