/* Embedded PC's ISA Bus Chapter 9 -- Ed Nisley                              */
/* Micro-C program to exercise DS2400 Silicon Serial Number                  */
/* You =must= remove the system Reset connector and -IOCHCHK jumper          */
/*  these tests do not twiddle the watchdog timer output bit                 */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

/*---------------------------------------------------------------------------*/
/* DS2400 constants                                                          */

#define TIMESLOT  712                  /* time slot in ticks:  99 us = 719   */
#define TIME_ONE  40                   /* low time for 1 bit:   5 us = 40    */
#define TIME_ZERO 647                  /* low time for 0 bit:  90 us = 647   */
#define TIME_READ 40                   /* low time for read:    5 us = 40    */
#define TIME_RD  15                    /* read delay in ticks:  2 us = 15    */

#define TIME_RESET   4315              /* reset pulse:        600 us = 4315  */
#define TIME_PDLOW   432               /*  presence low:       60 us = 432   */

/*--- 8254 channel and commands                                              */

#define DS_CTR    0                    /* counter channel                    */
#define DS_MODE   $30                  /* set channel to 16-bit, mode 0      */
#define DS_LATCH  $e2                  /* latch and read back status         */
#define DS_PIN    $80                  /* output pin bit in status value     */


/*---------------------------------------------------------------------------*/
/* Reset the DS2400                                                          */
/* Returns presence bit                                                      */
/*   0 = present, 1 = not present                                            */
/* Input bit must go high after reset pulse, then low                        */

int ResetDS2400()
{

int RetValue;

asm {

   CLI                                 ; need precise timing for this...

   MOV   DX,#SYNC_ADDR_A               ; show sync on printer port
   IN    AL,DX
   OR    AL,#$04
   OUT   DX,AL

   MOV   DX,#I8254_BASE_A+6            ; set mode
   MOV   AL,#DS_MODE
   OUT   DX,AL

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set initial high duration
   MOV   AX,#TIME_RESET
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   MOV   DX,#CTLS_ADDR_A               ; set DS2400 data bit high
   MOV   AX,#DS2400OUT_A
   OUT   DX,AX

   CALL  WaitTimer

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set reset pulse low duration
   MOV   AX,#TIME_RESET
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   MOV   DX,#CTLS_ADDR_A               ; set DS2400 data bit low
   MOV   AX,#0
   OUT   DX,AX

   CALL  WaitTimer

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set read delay interval
   MOV   AX,#TIME_RD
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   MOV   DX,#CTLS_ADDR_A               ; set DS2400 data bit high = floating
   MOV   AX,#DS2400OUT_A
   OUT   DX,AX

   CALL  WaitTimer

   MOV   DX,#STAT_ADDR_A               ; read output from DS2400
   IN    AX,DX
   MOV   BX,AX

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set delay to sample presence output
   MOV   AX,#TIME_PDLOW
   SUB   AX,#TIME_RD                   ; minus time we spent reading
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   CALL  WaitTimer

   MOV   DX,#STAT_ADDR_A               ; read output from DS2400
   IN    AX,DX
   MOV   CX,AX

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set remainder of reset time
   MOV   AX,#TIME_RESET
   SUB   AX,#TIME_PDLOW
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   STI                                 ; timing is not critical here

   CALL  WaitTimer

   MOV   DX,#SYNC_ADDR_A               ; clear sync pulse
   IN    AL,DX
   XOR   AL,#$04
   OUT   DX,AL

   XOR   CX,BX                         ; sampled bits must differ
   AND   CX,BX                         ; and first sample must be high

   MOV   AX,#0                         ; assume DS2400 is present
   TEST  CX,#DS2400IN_A                ;  ... should be 1
   JNZ   ResetDone
   INC   AX                            ; not present, return 1

ResetDone EQU *
   MOV   -2[BP],AX

}

   return RetValue;

}


/*---------------------------------------------------------------------------*/
/* Pulse the DS2400 output port bit                                          */
/* A pulse starts high, goes low for LowTime, then returns high for HighTime */
/* Times are measured in Firmware Development Board timer ticks: 139 ns      */
/* HighTime must be >=2 us (15 ticks) to allow data read delay               */
/* There is a mandatory 1 us delay added to each HighTime                    */
/* Returns the input state just after the start of HighTime                  */
/*  ... which is the data bit if the pulse marks a read time slot            */
/* We monopolize a counter and use Mode 0 to generate the time delay         */
/* Interrupts must be OFF for the first part of the time slot...             */

int PulseDS2400(LowTime,HighTime)
WORD LowTime;
WORD HighTime;
{

WORD TestValue;
int RetValue;

   TestValue;                          /* keep the compiler happy...         */
   HighTime;
   LowTime;

asm {
   CLI                                 ; need precise timing for first pulse

   MOV   DX,#SYNC_ADDR_A               ; show sync on printer port
   IN    AL,DX
   OR    AL,#$02
   OUT   DX,AL

   MOV   DX,#I8254_BASE_A+6            ; set mode
   MOV   AL,#DS_MODE
   OUT   DX,AL

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set low time
   MOV   AX,6[BP]
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   MOV   DX,#CTLS_ADDR_A               ; set DS2400 data bit low
   MOV   AX,#0
   OUT   DX,AX

   CALL  WaitTimer

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set read delay interval
   MOV   AX,#TIME_RD
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   MOV   DX,#CTLS_ADDR_A               ; set DS2400 data bit high = floating
   MOV   AX,#DS2400OUT_A
   OUT   DX,AX

   CALL  WaitTimer

   MOV   DX,#STAT_ADDR_A               ; read output from DS2400
   IN    AX,DX
   AND   AX,#DS2400IN_A                ; isolate the data bit
   MOV   -2[BP],AX

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; set rest of high time
   MOV   AX,4[BP]
   SUB   AX,#TIME_RD
   ADD   AX,#7                         ;  ... plus mandatory recovery time
   OUT   DX,AL
   XCHG  AH,AL
   OUT   DX,AL

   STI                                 ; re-enable interrupts

   CALL  WaitTimer

   JMP   PulseDone

;--- wait for timer interval to expire

WaitTimer EQU *
   MOV   DX,#I8254_BASE_A+6            ; latch & read back counter 0 status
   MOV   AL,#DS_LATCH
   OUT   DX,AL

   MOV   DX,#I8254_BASE_A+(DS_CTR*2)   ; read counter status
   IN    AL,DX

   TEST  AL,#DS_PIN                    ; wait for pin to go high again
   JZ    WaitTimer
   RET

PulseDone EQU *

   MOV   DX,#SYNC_ADDR_A               ; clear sync pulse
   IN    AL,DX
   XOR   AL,#$02
   OUT   DX,AL


}

   return !!RetValue;                              /* return Boolean value   */

}

/*---------------------------------------------------------------------------*/
/* Write a byte to the DS2400                                                */
/* Bits go out low-order first, which is not documented in the data sheet!   */

WrDS2400(Value)
WORD Value;
{

int BitCount;

   for (BitCount=0; BitCount < 8; ++BitCount) {
      if (Value & 0x0001) {                                 /* bit = 1       */
         PulseDS2400(TIME_ONE,TIMESLOT-TIME_ONE);
      }
      else {
         PulseDS2400(TIME_ZERO,TIMESLOT-TIME_ZERO);         /* bit = 0       */
      }
      Value >>= 1;
   }

   return;

}

/*---------------------------------------------------------------------------*/
/* Read a byte from the DS2400                                               */
/* Bits come in low-order first, too...                                      */

int RdDS2400()
{

int BitCount;
int RetValue;

   RetValue = 0;                                   /* clear return value     */

   for (BitCount=0; BitCount < 8; ++BitCount) {
      RetValue >>= 1;
      RetValue |= PulseDS2400(TIME_READ,TIMESLOT-TIME_READ) ? 0x0080 : 0;
   }

   return RetValue;

}


/*---------------------------------------------------------------------------*/
/* Compute DS2400 CRC value on complete data string                          */
/*  pData points to the 01 type code, follwed by six data bytes              */
/*  CRC is returned in the low-order byte                                    */
/* Remember that the 8051 and 8086 instruction sets are similar...           */
/*  but things like XOR and XRL set the flags differently!                   */

int CRCDS2400(pData)
BYTE *pData;
{

int CRC;

   pData = pData;                      /* keep compiler happy                */
   CRC = 0;                            /* clear both bytes                   */

asm {
   MOV   AH,#0                   ; set up initial CRC
   MOV   SI,4[BP]                ; set up data pointer
   MOV   CX,#7                   ; set up byte counter
OuterLoop EQU *
   PUSH  CX                      ; save byte counter
   MOV   CX,#8                   ; set up bit counter
   MOV   BL,[SI]                 ; fetch data byte
CRCLoop EQU *
   MOV   AL,BL                   ; set up data byte
   XOR   AL,AH                   ; combine low bit with old CRC
   ROR   AX,1                    ;  ... put result into high bit of CRC
   JNC   Zero                    ;  ... and into carry flag, too
   XOR   AH,#$0C                 ; stir in constant factor (pre-shifted!)
Zero EQU *
   SHR   BL,1                    ; set up next data bit
   LOOP  CRCLoop                 ; repeat for all bits
   INC   SI                      ; aim at next byte
   POP   CX                      ; fetch byte counter
   LOOP  OuterLoop               ; and repeat for each byte
   MOV   -2[BP],AH               ; set up return value
}

   return CRC;

}


/*---------------------------------------------------------------------------*/
/* The main event                                                            */

char Option;
WORD Counter;
BYTE SerNum[8];                        /* data from DS2400                   */

main() {

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("\nDS2400 serial number tests\n");
   putstr("Embedded PC's ISA Bus Chapter 9 -- Ed Nisley\n");

   while (TRUE) {

      putstr("\n\nHardware checkout routines:\n");
      putstr("== Remove System Reset connection and -IOCHCHK jumper! ==\n");
      putstr(" 1 - Write 0x0F command to DS2400, do not read back\n");
      putstr(" 2 - Read and display DS2400 data in loop\n");
      putstr("Select option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '1' :
         putstr("Observe DS2400 data line for 0x0F command value\n");
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            WrDS2400(0x0F);
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(HI(Counter)));       /* use high byte */
            if (BIOSWait(0x0000,0x2710)) {         /* delay for 10 ms        */
               putstr("BIOS delay is broken!\n");
               break;
            }
            ++Counter;
         }
         break;
      case '2' :
         while (!chkch()) {
            putstr("Read DS2400...");
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
            if (ResetDS2400()) {                   /* reset & check it       */
               putstr(" NOT");
            }
            putstr(" installed,");
            WrDS2400(0x0F);                        /* send command           */
            for (Counter=0; Counter < sizeof(SerNum); ++Counter) {
               SerNum[Counter] = RdDS2400();       /* read ID, data, & CRC   */
            }
            outp(SYNC_ADDR,0x00);
            printf(" type=%02x,",SerNum[0]);
            putstr(" data=");
            for (Counter=1; Counter < 7; ++Counter) {
               printf(" %02x",SerNum[Counter]);
            }
            printf(", CRC=%02x",SerNum[7]);
            Counter = CRCDS2400(SerNum);
            if (SerNum[7] == Counter) {
               putstr(" OK\n");
            }
            else {
               printf(" ** %02x\n",Counter);
            }
            if (BIOSWait(0x0001,0x86A0)) {         /* delay for 100 ms       */
               putstr("BIOS delay is broken!\n");
               break;
            }
         }
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
