/* Embedded PC's ISA Bus Chapter 10 -- Ed Nisley                             */
/* Micro-C program to demonstrate BIOS extensions                            */
/* REQUIRES 8086RLXS.ASM startup code in place of 8086RLPS.ASM               */
/*  ... modify SMALL.LIB or use BiosExt.LIB with MCCOMP.BAT                  */
/* You =must= remove the system Reset connector and -IOCHCHK jumper, because */
/*  these tests do not twiddle the watchdog timer output bit                 */

#include <8086io.h>
#include <8086int1.h>
#include "8086base.h"
#include "firmdev.h"

/*---------------------------------------------------------------------------*/
/* Global variables                                                          */

WORD Int18Off;                         /* previous Int 18 handler            */
WORD Int18Seg;

WORD Int19Off;                         /* previous Int 19 handler            */
WORD Int19Seg;

WORD Int61Off;                         /* previous Int 61 handler            */
WORD Int61Seg;

WORD StkOff;                           /* pointer to RAM and stack segment   */
WORD StkSeg;

/*---------------------------------------------------------------------------*/
/* Int 18 handler gets control just after the disk boot fails                */
/* This will not be called if the diskette or hard disk is bootable          */
/* This must =not= return, because there is no boot disk present!            */

INT_SMALL(Int18,0) Int18Handler() {

int Counter;

   putstr("Interrupt 18 (boot fail) handler\n");
   outp(SYNC_ADDR,0x04);

   putstr(" entering spin loop... your code goes here!\n");
   enable();                           /* allow timer ticks                  */

   while (1) {
      outpw(LED_ADDR,~ByteToSegs(Counter>>8));
      ++Counter;
   }

}


/*---------------------------------------------------------------------------*/
/* Int 19 handler gets control just before the disk boot attempt             */
/* We perform our trick and return to the normal BIOS boot routine           */
/* This discard all the registers saved across both stacks                   */

INT_SMALL(Int19,0) Int19Handler() {

int Option;

   serinit(9600,1);

   enable();                           /* allow timer ticks                  */

   putch(FORMFEED);
   putstr("Embedded PC's ISA Bus Chapter 10 -- Ed Nisley\n");
   putstr("Micro-C BIOS Extension Demo\n\n");
   putstr("Interrupt 19 (pre-boot) handler\n");

   outp(SYNC_ADDR,0x02);

/*--- show that we have got proper memory addressing                         */

   StkSeg = peekw(NV_SEGMENT,DATASEGMENT);
   StkOff = peekw(NV_SEGMENT,STACKTOPOFF);
   printf(" Pointer in NV RAM to top of stack is %04x:%04x\n",
          StkSeg,StkOff);

/*--- capture Int 18 vector to route control to our handler                  */

   GetVector(0x18,&Int18Seg,&Int18Off);
   printf(" Capturing Int 18 vector, was %04x:%04x\n",
          Int18Seg,Int18Off);
   SetVector(0x18,GetCodeSeg(),INT_S_ENTRY(Int18Handler));
   outpw(CTLS_ADDR,NV_WENABLE);
   pokew(NV_SEGMENT,OLDINT18OFF,Int18Off);
   pokew(NV_SEGMENT,OLDINT18SEG,Int18Seg);
   outpw(CTLS_ADDR,0);

/*--- see if Int 61 survived the BIOS boot process                           */

   GetVector(0x61,&Int61Seg,&Int61Off);
   if (Int61Seg != peekw(NV_SEGMENT,OLDINT19SEG)) {
      putstr("*** BIOS overwrote our info in the Int 61 vector!\n");
      printf("    we wrote %04x:%04x\n",
              peekw(NV_SEGMENT,OLDINT19SEG),
              peekw(NV_SEGMENT,OLDINT19OFF));
      printf("    is now   %04x:%04x\n",Int61Seg,Int61Off);
      SetVector(0x61,peekw(NV_SEGMENT,OLDINT19SEG),
                     peekw(NV_SEGMENT,OLDINT19OFF));
      putstr("    Restored to proper value!\n");
   }

/*--- now run the command loop                                               */

   while (1) {
      putstr(" Press Enter to reboot or Esc to invoke Int 18h: ");
      Option = getch();
      putch('\n');
      switch (Option) {
      case '\n' :
         printf(" invoking old Int 19 through Int 61 -> %04x:%04x...\n",
                 Int19Seg,Int19Off);
asm {
         INT   $61
}
         break;
      case ESC :
         printf(" invoking our Int 18 handler...\n");
asm {
         INT   $18
}
         break;
      }
   }


}


/*---------------------------------------------------------------------------*/
/* The main event                                                            */
/* This is invoked as part of the BIOS power-on reset sequence               */
/*  ... the startup code converts the BIOS extension call to our format      */
/* This routine =must= exit to allow the boot sequence to continue           */
/* The serial ports are not set up, so we cannot say anything just yet...    */

main() {

   GetVector(0x19,&Int19Seg,&Int19Off);            /* move old handler       */
   SetVector(0x61,Int19Seg,Int19Off);              /*  ... to new interrupt  */
   outpw(CTLS_ADDR,NV_WENABLE);                    /* save in NV RAM, also   */
   pokew(NV_SEGMENT,OLDINT19OFF,Int19Off);
   pokew(NV_SEGMENT,OLDINT19SEG,Int19Seg);
   outpw(CTLS_ADDR,0);

   SetVector(0x19,GetCodeSeg(),INT_S_ENTRY(Int19Handler));

   outp(SYNC_ADDR,0x01);

   return;


}


