/* Embedded PC's ISA Bus Chapters 12-14 -- Ed Nisley                                         */
/* Micro-C program to exercise Firmware Development Board Graphics LCD Intf  */

#include <8086io.h>
#include "8086base.h"
#include "firmdev.h"

#define TRACINGx                 /* define TRACING to dump progress reports  */
#define BLINKING                 /* define BLINKING to enable blinking       */
                                 /*  ... for those panels that support it    */

#define MAX_ERRORS   10          /* stop RAM tests after a few mismatches    */

char Option;
WORD Counter;
WORD RAMSeg;
WORD RAMAddr;
WORD RAMSize;
BYTE TestData;
int  ErrorCounter;

BYTE PRSData;
BYTE *pPRSData;
WORD PRSSegment;

int  Simple;
WORD NumRows;
WORD NumCols;
WORD Row,Col;
WORD Align;

char DigitString[100];

/*---------------------------------------------------------------------------*/
/* Utility routines for fetching & setting packed bytes                      */
/* Align is 0 for low nybble, 4 for high nybble                              */
/* Data is passed to and from these routines in the low nybble               */

WORD FetchNybble(Addr,Align)
int  Addr;
WORD Align;
{
WORD Temp;

   Temp = 0x0f & (peek(GLCD_SEGMENT,Addr) >> Align);

   return Temp;

}

StoreNybble(Addr,Align,Data)
WORD Addr;
int  Align;
WORD Data;
{
WORD OldData;

   OldData = peek(GLCD_SEGMENT,Addr) & (0xf0 >> Align);

   Data = ((Data & 0x0f) << Align) | OldData;

#ifdef TRACING
   printf(" s=%04x:%02x/%02x\r",OldData,Data);
#endif

   poke(GLCD_SEGMENT,Addr,Data);

}


/*---------------------------------------------------------------------------*/
/* Utility routine to set blinking values in high nybble                     */
/* Blink is 0 for no blinking (duplicated nybbles)                           */
/*         1 for foreground blink (high nybble = 0)                          */
/*         2 for background blink (high nybble = F)                          */
/*         3 for inverse blink (high nybble = complement of low nybble)      */

#define BLINK_OFF    0
#define BLINK_FG     1
#define BLINK_BG     2
#define BLINK_INV    3

SetData(Addr,Blink,Data)
WORD Addr;
int  Blink;
WORD Data;
{

   Data &= 0x000F;                                 /* force high nybble off  */

   switch (Blink) {
   default :
      printf("Invalid blink mode = %u\n",Blink);
   case BLINK_OFF :
      Data |= Data << 4;
      break;
   case BLINK_FG :
      break;
   case BLINK_BG :
      Data |= 0x00F0;
      break;
   case BLINK_INV :
      Data |= (~Data) << 4;
      break;
   }

   poke(GLCD_SEGMENT,Addr,Data);

}



/*---------------------------------------------------------------------------*/
/* Utility routines to handle LM215 packed bits                              */
/* Row and Col specify the dot within the overall 480x128 array              */
/* Align is 0..3 for bits 0..3                                               */
/* Blink is 0 for no blinking (duplicated nybbles)                           */
/*         1 for foreground blink (blink is OFF)                             */
/*         2 for background blink (blink is ON)                              */
/*         3 for inverse blink (blink to inverse of data bit)                */
/* Assumes data is duplicated in successive even and odd bytes               */

WORD FetchBit(Row,Col)
WORD Row;
WORD Col;
{
WORD Temp;
WORD RAMAddr;
WORD Align;

   RAMAddr = 2 * (240 * (Row % 64) + (Col % 240)) + 1;

   Align = (Col > 239) ? 2 : 0 + (Row > 63) ? 1 : 0;

   Temp = peek(GLCD_SEGMENT,RAMAddr);              /* fetch the bits         */
   Temp = (Temp >> Align) & 0x0001;                /* align & isolate        */

   return Temp;

}


StoreBit(Row,Col,Blink,Data)
WORD Row;
WORD Col;
WORD Blink;
WORD Data;
{
WORD Temp;
WORD Mask;
WORD RAMAddr;
WORD Align;

   RAMAddr = 2 * (240 * (Row % 64) + (Col % 240)); /* even address           */

   Align = ((Col > 239) ? 2 : 0) + ((Row > 63) ? 1 : 0);
   Mask = 0x0001 << Align;
   Data &= 0x0001;

   Temp = peek(GLCD_SEGMENT,RAMAddr + 1);          /* fetch the old bits     */
   Temp &= ~(Mask | (Mask << 4));                  /* strip target & blink   */
   Temp |= Data << Align;                          /* insert new data bit    */

   switch (Blink) {
   default :
      putstr("Invalid blink mode in Storebit: %u\n",Blink);
   case BLINK_OFF :
      Temp |= Data << (Align + 4);     /* no blink = duplicate the data bit  */
      break;
   case BLINK_FG :                     /* zero the blink bit, already done   */
      break;
   case BLINK_BG :
      Temp |= Mask << 4;               /* set the blink bit                  */
      break;
   case BLINK_INV :
      Temp |= ((~Data) & Mask) << 4;               /* inverse of data bit    */
      break;
   }

   poke(GLCD_SEGMENT,RAMAddr,Temp);
   poke(GLCD_SEGMENT,RAMAddr+1,Temp);

}

StoreByte(Row,Col,Blink,Data)
WORD Row;
WORD Col;
WORD Blink;
WORD Data;
{

   StoreBit(Row,Col  ,Blink,Data >> 7);
   StoreBit(Row,Col+1,Blink,Data >> 6);
   StoreBit(Row,Col+2,Blink,Data >> 5);
   StoreBit(Row,Col+3,Blink,Data >> 4);
   StoreBit(Row,Col+4,Blink,Data >> 3);
   StoreBit(Row,Col+5,Blink,Data >> 2);
   StoreBit(Row,Col+6,Blink,Data >> 1);
   StoreBit(Row,Col+7,Blink,Data     );

}

/*---------------------------------------------------------------------------*/
/* The main event...                                                         */

main() {

   serinit(9600,1);
   enable();

   putch(FORMFEED);
   putstr("\nGraphics LCD Interface Tests\n");
   putstr("Embedded PC's ISA Bus Chapters 12-44 -- Ed Nisley\n");

   outp(CTLS_ADDR,0);                  /* enable LCD buffer access           */

asm {
   MOV   >RANDSEED,#$0001
}
   PRSSegment = GetDataSeg() + 0x1000;             /* step 64K above data    */
   Simple = 0;

   while (TRUE) {

      putstr("\n\nGraphic LCD checkout routines:\n");
      printf(" 0 - Toggle Simple test mode (now %s)\n",Simple ? "ON" : "OFF");
      putstr(" 1 - Exercise LCD control latch\n");
      putstr(" 2 - Start 82C54 and Dot Clocks\n");
      putstr(" 3 - Repeated LCD buffer read\n");
      putstr(" 4 - Repeated LCD buffer write\n");
      putstr(" 5 - Verify LCD buffer with pseudorandom data\n");
      putstr(" 6 - Test: 640x200 4-bit 480 ns (160 clocks x 200 lines)\n");
      putstr(" 7 - Test: 640x200 4-bit 480 ns (320 clocks x 100 lines)\n");
      putstr(" 8 - Test: 640x400 8-bit 480 ns (160 clocks x 200 lines)\n");
      putstr(" 9 - Test: 640x400 4-bit 240 ns (320 clocks x 200 lines)\n");
      putstr(" A - Test: 640x480 4-bit 240 ns (320 clocks x 240 lines)\n");
      putstr(" B - Test: 480x128 4-bit 960 ns (240 clocks x 64  lines)\n");
      putstr("\nSelect option: ");

      Option = getch();
      putch(Option);
      putstr("    (press any key to exit test loops)\n\n");

      switch (Option) {
      case '0' :
         Simple = !Simple;
         printf("Simple mode is now %s\n",Simple ? "ON" : "OFF");
         break;
      case '1' :
         putstr("Verify walking 1 bit in LCD control latch... ");
         while (!chkch()) {
            for (TestData = 0x01;
                 TestData;
                 TestData = 0x00FF & (TestData << 1)) {
               outpw(GLCD_CTLS,TestData);
            }
         }
         break;
      case '2' :
         putstr("Clocks per LCD row (default 160): ");
         getstr(DigitString,sizeof(DigitString)-1);
         NumCols = DigitString[0] ? atoi(DigitString) : 160;
         putstr("Clocks per LCD frame (default 32000): ");
         getstr(DigitString,sizeof(DigitString)-1);
         NumRows = DigitString[0] ? atoi(DigitString) : 32000;
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,NumCols,I8254_BASE);     /* mode 2 binary          */
         LoadTimer(2,0x04,NumRows,I8254_BASE);     /* mode 2 binary          */
         outpw(GLCD_CTLS,GLCD_ON);                 /* enable clocking        */
         putstr(" ... verify 82C54, Dot Clocks, and sync pulses\n");
         break;
      case '3' :
         RAMSize = GLCD_SIZE;                      /* set up for assembler   */
         RAMSeg = GLCD_SEGMENT;
         printf("Reading memory at segment %04x, size %04x\n",
                RAMSeg,RAMSize);
         Counter = 0;
         while (!chkch()) {
            outp(SYNC_ADDR,0x01);                  /* sync on printer port   */
asm {
   MOV   CX,RAMSize        repeat for whole block
   PUSH  DS
   MOV   AX,RAMSeg         set up RAM address
   MOV   DS,AX
   XOR   SI,SI
   REP
   LODSB
   POP   DS
}
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(Counter));
            ++Counter;
         }
         break;
      case '4' :
         printf("Writing memory at segment %04x, size %04x\n",
                GLCD_SEGMENT,GLCD_SIZE);
         RAMAddr = 0x0000;
         Counter = 0;
         while (!chkch()) {
            if (RAMAddr == GLCD_SIZE) {
               outp(SYNC_ADDR,0x01);               /* sync on printer port   */
               outpw(LED_ADDR,~ByteToSegs(Counter));
               ++Counter;
               RAMAddr = 0;
            }
            outp(SYNC_ADDR,0x02);                  /* mark each peek         */
            poke(GLCD_SEGMENT,RAMAddr,~RAMAddr);   /* write inverse address  */
            outp(SYNC_ADDR,0x00);
            ++RAMAddr;
         }
         break;
      case '5' :
         putstr("Calculating pseudorandom data...  ");
         for (pPRSData = 0; pPRSData != 0x8000; pPRSData += 2) {
            pokew(PRSSegment,pPRSData,rand(0xFFFF));
         }
         putstr("done\n");
         printf("Loading 32K RAM at seg %04x... ",GLCD_SEGMENT);
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != GLCD_SIZE; ++RAMAddr) {
            PRSData = peek(PRSSegment,RAMAddr);
            outp(SYNC_ADDR,0x01);                  /* mark the write         */
            poke(GLCD_SEGMENT,RAMAddr,PRSData);    /* do it                  */
            TestData = peek(GLCD_SEGMENT,RAMAddr);
            if (TestData != PRSData) {
               printf("Cannot set %04x to %02x, is %02x\n",
                  RAMAddr,PRSData,TestData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
            outpw(LED_ADDR,~ByteToSegs(HI(RAMAddr)));
         }
         putstr("done\n");
         putstr("Verifying data... ");
         ErrorCounter = 0;
         for (RAMAddr = 0x0000; RAMAddr != GLCD_SIZE; ++RAMAddr) {
            outp(SYNC_ADDR,0x01);                  /* mark this loop         */
            TestData = peek(GLCD_SEGMENT,RAMAddr);
            PRSData = peek(PRSSegment,RAMAddr);
            if (PRSData != TestData) {
               printf("Mismatch at %04x, data is %02x, should be %02x\n",
                       RAMAddr,TestData,PRSData);
               if (MAX_ERRORS < ++ErrorCounter) {
                  putstr("Too many errors, giving up\n");
                  break;
               }
            }
            outp(SYNC_ADDR,0x00);
         }
         putstr("done\n");
         break;
      case '6' :
         putstr("640 x 200 dots, 4 bits, 480 ns clock, blinking\n");
         putstr("  - display has 200 row drivers\n");
         putstr("  - clock jumper to +Dot Clock\n");
         putstr(" Optrex DMF651\n");
         putstr(" ... setting counter");
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,160,I8254_BASE);         /* clocks per row         */
         LoadTimer(2,0x04,32000,I8254_BASE);       /* total clocks/frame     */
         putstr(" ... medium blink rate\n");
#ifdef BLINKING
         outpw(GLCD_CTLS,GLCD_ON + GLCD_BLMED);    /* start and set blinking */
#else
         outpw(GLCD_CTLS,GLCD_ON);                 /* no blinking for photo! */
#endif
         putstr(" ... test data: row number, address, diagonal stripe\n");
         for (Row = 0; Row < 200; Row++) {
            RAMAddr = Row * 160;
            for (Col = 0; Col < 160; Col++) {
               poke(GLCD_SEGMENT,RAMAddr+Col,0x00);         /* clear row     */
            }
            SetData(RAMAddr+0,BLINK_OFF,Row >> 4);
            SetData(RAMAddr+1,BLINK_OFF,Row);
            SetData(RAMAddr+2,BLINK_FG,0x09);      /* divider                */
            SetData(RAMAddr+3,BLINK_OFF,RAMAddr >> 12);
            SetData(RAMAddr+4,BLINK_OFF,RAMAddr >> 8);
            SetData(RAMAddr+5,BLINK_OFF,RAMAddr >> 4);
            SetData(RAMAddr+6,BLINK_OFF,RAMAddr);
            SetData(RAMAddr+7,BLINK_FG,0x09);     /* divider                */
            SetData(RAMAddr+8+(Row >> 2),BLINK_BG,0x08 >> (Row & 0x03));
         }
         putstr(" ... done\n");
         break;
      case '7' :
         putstr("640 x 200 dots, 4 bits, 480 ns clock, blinking\n");
         putstr("  - display has 100 row drivers\n");
         putstr("  - clock jumper to +Dot Clock\n");
         putstr(" Toshiba TLY-365-121\n");
         putstr(" ... setting counter");
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,320,I8254_BASE);         /* clocks per row         */
         LoadTimer(2,0x04,32000,I8254_BASE);       /* total clocks/frame     */
         putstr(" ... medium blink rate\n");
#ifdef BLINKING
         outpw(GLCD_CTLS,GLCD_ON + GLCD_BLMED);    /* start and set blinking */
#else
         outpw(GLCD_CTLS,GLCD_ON);                 /* no blinking for photo! */
#endif
         putstr(" ... test data: row number, address, diagonal stripe\n");
         for (Row = 0; Row < 200; Row++) {
            RAMAddr = 320 * (Row % 100) + ((Row > 99) ? 160 : 0);
            for (Col = 0; Col < 160; Col++) {
               poke(GLCD_SEGMENT,RAMAddr+Col,0x00);         /* clear row     */
            }
            SetData(RAMAddr+0,BLINK_OFF,Row >> 4);
            SetData(RAMAddr+1,BLINK_OFF,Row);
            SetData(RAMAddr+2,BLINK_FG,0x09);      /* divider                */
            SetData(RAMAddr+3,BLINK_OFF,RAMAddr >> 12);
            SetData(RAMAddr+4,BLINK_OFF,RAMAddr >> 8);
            SetData(RAMAddr+5,BLINK_OFF,RAMAddr >> 4);
            SetData(RAMAddr+6,BLINK_OFF,RAMAddr);
            SetData(RAMAddr+7,BLINK_FG,0x09);     /* divider                */
            SetData(RAMAddr+8+(Row >> 2),BLINK_BG,0x08 >> (Row & 0x03));
         }
         putstr(" ... done\n");
         break;
      case '8' :
         putstr("640 x 400 dots, 8 bits, 480 ns clock\n");
         putstr("  - clock jumper to +Dot Clock\n");
         putstr(" Matsushita EDM-LG64AA44D\n");
         putstr(" ... setting counter");
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,160,I8254_BASE);         /* clocks per row         */
         LoadTimer(2,0x04,32000,I8254_BASE);       /* total clocks/frame     */
         putstr(" ... no blinking\n");
         outpw(GLCD_CTLS,GLCD_ON | GLCD_BLZERO);   /* get bits 0-3 from mux  */
         putstr(" ... clearing buffer\n");
         for (Row = 0; Row < 400; Row++) {
            RAMAddr = 160 * (Row % 200);           /* buffer address         */
            Align = (Row >= 200) ? 4 : 0;          /* high/low bit alignment */
            for (Col = 0; Col < 160; Col++) {
               StoreNybble(RAMAddr+Col,Align,0);   /* clear our nybble       */
            }
         }
         if (Simple) {
            putstr("Simple test data\n");
            poke(GLCD_SEGMENT,0x0000,0x0f);
            poke(GLCD_SEGMENT,0x0001,0x00);
            poke(GLCD_SEGMENT,0x0002,0x0f);
            poke(GLCD_SEGMENT,0x0003,0x00);
            poke(GLCD_SEGMENT,0x0004,0x0e);
            poke(GLCD_SEGMENT,0x0005,0x00);
            poke(GLCD_SEGMENT,0x0006,0x0e);
            poke(GLCD_SEGMENT,0x0007,0x00);
            poke(GLCD_SEGMENT,0x0008,0x0c);
            poke(GLCD_SEGMENT,0x0009,0x00);
            poke(GLCD_SEGMENT,0x000a,0x0c);
            poke(GLCD_SEGMENT,0x000b,0x00);
            poke(GLCD_SEGMENT,0x000c,0x08);
            poke(GLCD_SEGMENT,0x000d,0x00);
            poke(GLCD_SEGMENT,0x000e,0x08);
            poke(GLCD_SEGMENT,0x000f,0x00);
            poke(GLCD_SEGMENT,0x0010,0x01);
            poke(GLCD_SEGMENT,0x0011,0x03);
            poke(GLCD_SEGMENT,0x0012,0x07);
            poke(GLCD_SEGMENT,0x0013,0x0f);
         }
         else {
            putstr(" ... test data: row number, address, diagonal stripe\n");
            for (Row = 0; Row < 400; Row++) {
               RAMAddr = 160 * (Row % 200);        /* buffer address         */
               Align = (Row >= 200) ? 4 : 0;       /* high/low bit alignment */
               StoreNybble(RAMAddr+0,Align,Row >> 12);
               StoreNybble(RAMAddr+1,Align,Row >> 8);
               StoreNybble(RAMAddr+2,Align,Row >> 4);
               StoreNybble(RAMAddr+3,Align,Row);
               StoreNybble(RAMAddr+4,Align,0x08);     /* divider lines       */
               StoreNybble(RAMAddr+5,Align,RAMAddr >> 12);
               StoreNybble(RAMAddr+6,Align,RAMAddr >> 8);
               StoreNybble(RAMAddr+7,Align,RAMAddr >> 4);
               StoreNybble(RAMAddr+8,Align,RAMAddr);
               StoreNybble(RAMAddr+9,Align,0x06);     /* divider lines       */
               StoreNybble(RAMAddr+10+(Row >> 2),Align,0x08 >> (Row & 0x03));
            }
         }
         putstr(" ... done\n");
         break;
      case '9' :
         putstr("640 x 400 dots, 4 bits, 240 ns clock\n");
         putstr("  - clock jumper to -2x Dot Clock\n");
         putstr("  - sync jumpers to +2x Frame & +2x Line Sync\n");
         putstr(" Sharp LM64015T\n");
         putstr(" ... setting counter");
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,160,I8254_BASE);         /* Dot Clocks/row         */
         LoadTimer(2,0x04,32000,I8254_BASE);       /* total clocks/frame     */
         putstr(" ... no blinking\n");
         outpw(GLCD_CTLS,GLCD_ON | GLCD_BLCLOCK);  /* alternate at clock rate*/
         putstr(" ... clearing buffer\n");
         for (Row = 0; Row < 400; Row++) {
            RAMAddr = 160 * (Row % 200) + ((Row > 199) ? 80 : 0);
            for (Col = 0; Col < 80; Col++) {
               poke(GLCD_SEGMENT,RAMAddr+Col,0xff);         /* 1 = clear     */
            }
         }
         if (Simple) {
            putstr("Simple test data\n");
            poke(GLCD_SEGMENT,0,~0xFF);         /* row 1                  */
            poke(GLCD_SEGMENT,1,~0xfe);
            poke(GLCD_SEGMENT,2,~0xef);
            poke(GLCD_SEGMENT,3,~0x18);
            poke(GLCD_SEGMENT,80,~0xc0);        /* row 200                */
            poke(GLCD_SEGMENT,81,~0x60);
            poke(GLCD_SEGMENT,82,~0x30);
            poke(GLCD_SEGMENT,83,~0x0f);
            poke(GLCD_SEGMENT,156,~0x55);       /* end of row 200         */
            poke(GLCD_SEGMENT,157,~0xaa);
            poke(GLCD_SEGMENT,158,~0xc0);
            poke(GLCD_SEGMENT,159,~0x03);
            poke(GLCD_SEGMENT,160,~0x80);       /* row 2                  */
            poke(GLCD_SEGMENT,161,~0x40);
            poke(GLCD_SEGMENT,162,~0x20);
            poke(GLCD_SEGMENT,163,~0x10);
         }
         else {
            putstr(" ... test data: row number, address, diagonal stripe\n");
            for (Row = 0; Row < 400; Row++) {
               RAMAddr = 160 * (Row % 200) + ((Row > 199) ? 80 : 0);
               poke(GLCD_SEGMENT,RAMAddr+0,~(Row >> 8));
               poke(GLCD_SEGMENT,RAMAddr+1,~(Row & 0x00FF));
               poke(GLCD_SEGMENT,RAMAddr+2,~0x99);    /* divider lines       */
               poke(GLCD_SEGMENT,RAMAddr+3,~(RAMAddr >> 8));
               poke(GLCD_SEGMENT,RAMAddr+4,~(RAMAddr & 0x00FF));
               poke(GLCD_SEGMENT,RAMAddr+5,~0x99);    /* divider lines       */
               poke(GLCD_SEGMENT,RAMAddr+6+(Row >> 3),
                                 ~(0x80 >> (Row & 0x0007)));
            }
         }
         putstr(" ... done\n");
         break;
      case 'a' :
      case 'A' :
         putstr("640 x 480 dots, 4 bits, 240 ns clock\n");
         putstr("  - requires 64 K byte RAM buffer\n");
         putstr("  - clock jumper to -2x Dot Clock\n");
         putstr("  - sync jumpers to +2x Frame & +2x Line Sync\n");
         putstr(" Sharp LM641481\n");
         putstr(" ... setting counter");
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,160,I8254_BASE);         /* Dot Clocks/row         */
         LoadTimer(2,0x04,38400,I8254_BASE);       /* total clocks/frame     */
         putstr(" ... no blinking\n");
         outpw(GLCD_CTLS,GLCD_ON | GLCD_BLCLOCK);  /* alternate at clock rate*/
         for (Row = 0; Row < 480; Row++) {
            RAMAddr = 160 * (Row % 240) + ((Row > 239) ? 80 : 0);
            for (Col = 0; Col < 80; Col++) {
               poke(GLCD_SEGMENT,RAMAddr+Col,0xff);         /* 1 = clear     */
            }
         }
         if (Simple) {
            putstr("Simple test data\n");
            poke(GLCD_SEGMENT,0,~0xFF);            /* row 1                  */
            poke(GLCD_SEGMENT,1,~0xfe);
            poke(GLCD_SEGMENT,2,~0xef);
            poke(GLCD_SEGMENT,3,~0x18);
            poke(GLCD_SEGMENT,160,~0x80);          /* row 2                  */
            poke(GLCD_SEGMENT,161,~0x40);
            poke(GLCD_SEGMENT,162,~0x20);
            poke(GLCD_SEGMENT,163,~0x10);
         }
         else {
            putstr(" ... test data: row number, address, diagonal stripe\n");
            for (Row = 0; Row < 480; Row++) {
               RAMAddr = 160 * (Row % 240) + ((Row > 239) ? 80 : 0);
               poke(GLCD_SEGMENT,RAMAddr+0,~(Row >> 8));
               poke(GLCD_SEGMENT,RAMAddr+1,~(Row & 0x00FF));
               poke(GLCD_SEGMENT,RAMAddr+2,~0x99);    /* divider lines       */
               poke(GLCD_SEGMENT,RAMAddr+3,~(RAMAddr >> 8));
               poke(GLCD_SEGMENT,RAMAddr+4,~(RAMAddr & 0x00FF));
               poke(GLCD_SEGMENT,RAMAddr+5,~0x99);    /* divider lines       */
               poke(GLCD_SEGMENT,RAMAddr+6+(Row >> 3),
                                  ~(0x80 >> (Row & 0x0007)));
            }
         }
         putstr(" ... done\n");
         break;
      case 'b' :
      case 'B' :
         putstr("480 x 128 dots, 4 bits, 960 ns clock\n");
         putstr("  - clock jumper to -Dot Clock/2\n");
         putstr(" Hitachi LM215\n");
         putstr(" ... setting counter");
         outpw(GLCD_CTLS,GLCD_OFF);                /* disable clocking       */
         LoadTimer(1,0x04,480,I8254_BASE);         /* Dot Clocks/row         */
         LoadTimer(2,0x04,30720,I8254_BASE);       /* total clocks/frame     */
         putstr(" ... fast blink rate\n");
         outpw(GLCD_CTLS,GLCD_ON | GLCD_BLFAST);   /* enable & set blinking  */
         putstr(" ... clearing buffer\n");
         for (Row = 0; Row < 64; Row++) {
            for (Col = 0; Col < 240; Col++) {
               RAMAddr = 2 * (240 * Row) + Col;
               poke(GLCD_SEGMENT,RAMAddr+Col,0x00);         /* even address  */
               poke(GLCD_SEGMENT,RAMAddr+Col+1,0x00);       /* odd address   */
            }
         }
         if (Simple) {
            putstr("Simple test data\n");
            poke(GLCD_SEGMENT,0,0x0F);             /* blinking dot/quadrant  */
            poke(GLCD_SEGMENT,1,0x0F);             /*  odd addresses only    */
            poke(GLCD_SEGMENT,2,0xFF);             /* non-blinking dot/quad  */
            poke(GLCD_SEGMENT,3,0xFF);
            poke(GLCD_SEGMENT,4,0x01);             /* UL only                */
            poke(GLCD_SEGMENT,5,0x01);
            poke(GLCD_SEGMENT,6,0x02);             /* LL only                */
            poke(GLCD_SEGMENT,7,0x02);
            poke(GLCD_SEGMENT,8,0x04);             /* UR only                */
            poke(GLCD_SEGMENT,9,0x04);
            poke(GLCD_SEGMENT,10,0x08);            /* LR only                */
            poke(GLCD_SEGMENT,11,0x08);
            poke(GLCD_SEGMENT,12,0xFF);            /* non-blinking dot/quad  */
            poke(GLCD_SEGMENT,13,0xFF);
            break;
         }
         else {
            putstr(" ... test data: row number, address, diagonal stripe\n");
            for (Row = 0; Row < 128; Row++) {
               RAMAddr = 2 * (240 * Row) + 1;
               StoreByte(Row,0,BLINK_OFF,Row);
               StoreByte(Row,8,BLINK_FG,0x81);
               StoreByte(Row,16,BLINK_OFF,RAMAddr >> 8);
               StoreByte(Row,24,BLINK_OFF,RAMAddr);
               StoreBit(Row,32+Row,BLINK_FG,1);

               StoreByte(Row,240+0,BLINK_OFF,Row);
               StoreByte(Row,240+8,BLINK_BG,0x7E);
               StoreByte(Row,240+16,BLINK_OFF,RAMAddr >> 8);
               StoreByte(Row,240+24,BLINK_OFF,RAMAddr);
               StoreBit(Row,240+32+Row,BLINK_FG,1);
            }
         }
         putstr(" ... done\n");
         break;
      default :
         putstr("*** Choose one of the menu options\n\n");
      }

   }


}
