/* Embedded PC's ISA Bus Chapter 15 -- Ed Nisley                             */
/* LCD Life routine                                                          */
/*  Displays Conway's Game of Life on Firmware Development Board's           */
/*  Graphic LCD Interface... =lots= of dots!                                 */
/* Stores LCD frames in extended memory for a slide show                     */
/* Compiled with Borland C++ 3.1, converted to binary with Paradim Locate    */
/* Requires diskette boot sector loader to 1000:0000                         */
/*  ... and a comm program on COM1 with CTS & DSR held high                  */

/*--- hardcoded constants                                                    */

#define DENSITY 2000                   /* initial dot density, 10000 = full  */
#define FRAMESIZE 32                   /* slide capture buffer in KB         */
#define MAXGENS 999                    /* maximum number of generations      */
#define REPLAYTIME 500000L             /* delay between slide show frames    */

/*--- switch settings to control operation                                   */

#define SIMPLE    0x8000               /* set to force simple test pattern   */
#define LOOPCOPY  0x4000               /* set to force LCD update by loop    */
#define EXITNOW   0x0100               /* set during creation to force exit  */

#define ONE_MEG (1024L * 1024L)        /* an easy way to enter a megabyte    */

#include <alloc.h>
#include <dos.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>

#include "8086base.h"
#include "firmdev.h"
#include "utility.h"
#include "console.h"

#include "LCDPlot.h"

/*--- define the GDT structure needed for extended memory block moves        */
/*    and a macro to convert a pointer into a linear 24 bit address          */

#pragma option -a-                     /* structure must be packed           */
typedef struct {
   WORD SegLimit;                      /* size of segment in bytes           */
   WORD SegBaseLow;                    /* low two bytes of linear address    */
   BYTE SegBaseHigh;                   /* high byte of linear address        */
   BYTE SegFlags;                      /* 9B = code, 93 = data, 91 = RO data */
   WORD Reserved;                      /* reserved, must be zero             */
} GDT_286;
#pragma option -a.                     /* restore default setting            */

#define MK_LINEAR(fp) (((long)FP_SEG(fp) << 4) + (long)FP_OFF(fp))

/*--- global variables, some shared with LCD driver routine                  */

int NumRows;
int NumCols;
unsigned int NumBytes;
BYTE far *pOldBuff;
BYTE far *pNewBuff;
BYTE far *pLCDBuff;
GDT_286 MoveGDT[6];                    /* GDT for BIOS Move Block operation  */


/*--- the main event...                                                      */

void  main(void)
{

WORD Generations;
int Row,Col;
int Neighbors;
WORD ByteOffset;
int RowOffset;
char *pName;
WORD Switches;
WORD XMemSize;
long XMemAddr;
long XMemLimit;
union REGS regs;
struct SREGS sregs;

   outpw(CTLS_ADDR,0);                 /* allow GLCD RAM buffer access       */

   _putch(FORMFEED);
   puts("Graphic LCD Interface: Game of Life");
   puts("Embedded PC's ISA Bus Chapter 15 -- Ed Nisley");

   puts("Initializing and clearing LCD panel...");
   LCDInit(&NumRows,&NumCols,&NumBytes,&pName);    /* start the LCD panel    */
   printf("  %s is now active \n"
          "  with %u rows and %u cols using %u bytes\n",
           pName,NumRows,NumCols,NumBytes);

   puts("Allocating buffers...");
   pOldBuff = farcalloc(NumBytes,1);               /* get Life buffers       */
   pNewBuff = pLCDBuff;                            /* new dots into LCD      */
   printf(" pLCDBuff=%Fp, pNewBuff=%Fp, pOldBuff=%Fp\n",
           pLCDBuff,pNewBuff,pOldBuff);

/*--- set up extended memory buffers to capture the frames                   */

   puts("Checking for extended memory...");
   regs.h.ah = 0x88;
   outp(SYNC_ADDR,0x01);                           /* mark the query         */
   int86(0x15,&regs,&regs);
   outp(SYNC_ADDR,0x00);
   XMemSize = regs.x.ax;
   printf(" BIOS reports %u KB\n",XMemSize);
   if ((XMemSize >> 8) >= 0x0080) {
      puts(" ... on a small PC you can't have that much, so it's ignored");
      XMemSize = 0;
   }
   XMemLimit = (long)(XMemSize / FRAMESIZE);       /* truncate limit to a    */
   XMemLimit *= (long)FRAMESIZE * 1024L;           /*  multiple of the slide */
   XMemLimit += ONE_MEG;                           /* above the line         */
   printf(" Room for %u frames of %u KB each between %08lX and %08lX\n",
         XMemSize/FRAMESIZE,FRAMESIZE,ONE_MEG,XMemLimit-1);


   Switches = ~inpw(SW_ADDR);                      /* fetch switches         */

   printf("Creating initial pattern...");
   srand(0x00FF & Switches);                       /* set up random numbers  */
   if (Switches & SIMPLE) {
      printf(" (diagonal)");
      for (Row=0; Row<NumRows; Row++) {            /* set up initial pattern */
         ByteToLEDs(Row);                          /* show some progress     */
         for (Col=0; Col<NumCols; Col++) {         /*  ... with the LCD ON   */
            if ((Row + Col) & 0x0001) {
               LCDSetDot(Row,Col,1);
            }
         }
      }
      puts(" press spacebar to continue...");
      _getch();
   }
   else {
      for (Row=0; Row<NumRows; Row++) {            /* set up initial pattern */
         ByteToLEDs(Row);                          /* show some progress     */
         for (Col=0; Col<NumCols; Col++) {         /*  ... with the LCD ON   */
            if (random(10000) < DENSITY) {
               LCDSetDot(Row,Col,1);
            }
         }
      }
   }

   _fmemcpy(pOldBuff,pNewBuff,NumBytes);           /* set up source buffer   */

   puts(" done");

/*--- generate the patterns and stash in extended memory for later           */
/*    if there's no extended memory, we'll never exit this loop              */

   Generations = 0;
   XMemAddr = ONE_MEG;                             /* store starting at 1MB  */

   while (1) {

      printf("\rGeneration ");

      if (Switches & SIMPLE) {
         for (Row=0; Row<NumRows; Row++) {
            RowOffset = (Generations % NumCols) - Row;
            ByteToLEDs(Row);                       /* show some progress     */
            for (Col=0; Col<NumCols; Col++) {
               if (RowOffset == Col) {
                  LCDSetDot(Row,Col,1);
               }
               else {
                  LCDSetDot(Row,Col,0);
               }
            }
         }
      }
      else {
         for (Row=0; Row<NumRows; Row++) {
            ByteToLEDs(Row);                    /* show some progress        */
            outp(SYNC_ADDR,0x80);               /* to time the calcuation    */
            for (Col=0; Col<NumCols; Col++) {
               Neighbors = LCDGetDot(Row-1,Col-1) +
                           LCDGetDot(Row-1,Col  ) +
                           LCDGetDot(Row-1,Col+1) +
                           LCDGetDot(Row  ,Col-1) +
                           LCDGetDot(Row  ,Col+1) +
                           LCDGetDot(Row+1,Col-1) +
                           LCDGetDot(Row+1,Col  ) +
                           LCDGetDot(Row+1,Col+1);
               if (LCDGetDot(Row,Col)) {
                  if ((Neighbors < 2) || (Neighbors > 3)) {
                     LCDSetDot(Row,Col,0);
                  }
               }
               else {
                  if (Neighbors == 3) {
                     LCDSetDot(Row,Col,1);
                  }
               }
            }
            outp(SYNC_ADDR,0x00);
         }
      }

      printf("%5u",Generations);

      _fmemcpy(pOldBuff,pNewBuff,NumBytes);        /* extract new dots       */

      if (XMemSize && (XMemAddr < XMemLimit)) {
         memset(MoveGDT,0,sizeof(MoveGDT));                 /* set up zeros  */
         MoveGDT[2].SegLimit = FRAMESIZE*1024;
         MoveGDT[2].SegFlags = 0x93;
         MoveGDT[2].SegBaseLow = (WORD)MK_LINEAR(pOldBuff);
         MoveGDT[2].SegBaseHigh = (BYTE)(MK_LINEAR(pOldBuff) >> 16);
         MoveGDT[3].SegLimit = FRAMESIZE*1024;
         MoveGDT[3].SegFlags = 0x93;
         MoveGDT[3].SegBaseLow = (WORD)XMemAddr;            /* target        */
         MoveGDT[3].SegBaseHigh = (BYTE)(XMemAddr >> 16);

         segread(&sregs);                          /* get current seg regs   */
         regs.h.ah = 0x87;                         /* move block             */
         regs.x.cx = FRAMESIZE*(1024/2);           /* number of words        */
         regs.x.si = FP_OFF(MoveGDT);              /* offset of GDT          */
         sregs.es = FP_SEG(MoveGDT);               /*  ... segment           */

         outp(SYNC_ADDR,0x02);                     /* mark the move          */
         int86x(0x15,&regs,&regs,&sregs);          /* shazam!                */
         outp(SYNC_ADDR,0x00);

         printf(" stored at %08lX",XMemAddr);

         XMemAddr += (long)FRAMESIZE * 1024L;      /* step to next block     */
      }

      ++Generations;
      Switches = ~inpw(SW_ADDR);                   /* allow early exit       */
      if ((Generations > MAXGENS) ||
          (XMemSize && (XMemAddr >= XMemLimit)) ||
          (Switches & EXITNOW)) {
         puts("\n... finished creating patterns");
         break;
      }

   }

/*--- now display the stored patterns fairly quickly                         */

   puts("Starting replay...");

   Generations = 0;
   XMemAddr = ONE_MEG;                             /* start at 1M line!      */

   while (1) {

      printf("\r Generation %5u",Generations);
      ByteToLEDs(Generations);

      memset(MoveGDT,0,sizeof(MoveGDT));                    /* set up zeros  */
      MoveGDT[2].SegLimit = FRAMESIZE*1024;
      MoveGDT[2].SegFlags = 0x93;
      MoveGDT[2].SegBaseLow = (WORD)XMemAddr;               /* source        */
      MoveGDT[2].SegBaseHigh = (BYTE)(XMemAddr >> 16);
      MoveGDT[3].SegLimit = FRAMESIZE*1024;
      MoveGDT[3].SegFlags = 0x93;
      MoveGDT[3].SegBaseLow = (WORD)MK_LINEAR(pOldBuff);
      MoveGDT[3].SegBaseHigh = (BYTE)(MK_LINEAR(pOldBuff) >> 16);

      segread(&sregs);                             /* get current seg regs   */
      regs.h.ah = 0x87;                            /* move block             */
      regs.x.cx = FRAMESIZE*(1024/2);              /* number of words        */
      regs.x.si = FP_OFF(MoveGDT);                 /* offset of GDT          */
      sregs.es = FP_SEG(MoveGDT);                  /*  ... segment           */

      outp(SYNC_ADDR,0x04);                        /* mark the move          */
      int86x(0x15,&regs,&regs,&sregs);             /* shazam!                */
      outp(SYNC_ADDR,0x00);

      printf(" fetched from %08lX",XMemAddr);

      Switches = ~inpw(SW_ADDR);                   /* allow in-flight tests  */
      if (Switches & LOOPCOPY) {
         for (ByteOffset = 0;
              ByteOffset < (WORD)(FRAMESIZE * 1024);
              ++ByteOffset) {
            *(pLCDBuff + ByteOffset) = *(pOldBuff + ByteOffset);
         }
      }
      else {
         _fmemcpy(pLCDBuff,pOldBuff,NumBytes);     /* blitz copy             */
      }

      XMemAddr += (long)FRAMESIZE * 1024L;         /* step to next block     */

      Generations++;

      if ((Generations > MAXGENS) ||
          (XMemAddr >= XMemLimit)) {
         XMemAddr = ONE_MEG;
         Generations = 0;
      }

      BIOSDelay(REPLAYTIME);                       /* allow a pause...       */

   }

}
