; Graphic LCD Interface routines for 640x400 8-bit 480 ns panel
; This has 200 rows of 160 clocks

              IDEAL
              P386N

              INCLUDE   "firmdev.inc"

              MODEL     SMALL,CPP

;-------------
; Define LCD parameters
; We must use = rather than EQU to get real numeric results!

LCDNAME       EQU       "Matsushita EDM LG64AA44D"
NUMROWS       =         400
NUMCOLS       =         640
COLCLOCKS     =         160
COLMODULUS    =         4
TOTALCLOCKS   =         32000
ROWMODULUS    =         NUMROWS/2
FILLVALUE     =         000h

              EXTRN     C Load8254:PROC

;-------------
; Set up data values needed by the plotting routines
; The 386 division routines need in-storage divisors...

              DATASEG

              EXTRN     C pOldBuff:FAR PTR BYTE
              EXTRN     C pNewBuff:FAR PTR BYTE
              EXTRN     C pLCDBuff:FAR PTR BYTE

LCDName       DB        LCDNAME             ; fill in the name
              DB        0                   ;  ... and a terminator
ColModulus    DW        COLMODULUS          ; convert dot col to byte offset

RowStarts     DW        NUMROWS DUP (?)     ; row starting addresses

LCDCtlBits    DW        0                   ; copy of GLCD_CTLS output port

              CODESEG

              EXTRN     _Load8254

;-------------
; Enable & disable the LCD panel

              PUBLIC    C LCDEnable
              PROC      LCDEnable
              ARG       Enable : WORD

              MOV       AX,[LCDCtlBits]     ; pick up current controls

              TEST      [Enable],0001h      ; 1 = turn it on...
              JNZ       @@On

              AND       AX,NOT GLCD_ENDISP  ; lower the enable line
              JMP       SHORT @@Set
@@On:
              OR        AX,GLCD_ENDISP      ; raise the enable line

@@Set:
              MOV       DX,GLCD_CTLS        ; tell the port
              OUT       DX,AX

              MOV       [LCDCtlBits],AX     ; and save for next time

              RET

              ENDP      LCDEnable


;-------------
; Initialize the Graphic LCD Interface hardware
; Also fills the buffer and returns some useful numbers
; This keeps the initialization code in here where it belongs...

              PUBLIC    C LCDInit
              PROC      LCDInit
              ARG       pNumRows : PTR WORD, pNumCols: PTR WORD, \
                        pNumBytes : WORD, ppName : PTR PTR BYTE
              USES      ES,DI

;--- disable the panel and set up the sync counters

              MOV       AL,GLCD_OFF         ; shut the interface down
              MOV       DX,GLCD_CTLS
              OUT       DX,AL

              CALL      Load8254 C,1,04h,COLCLOCKS    ; use Mode 2
              CALL      Load8254 C,2,04h,TOTALCLOCKS  ;  ... for both values

              MOV       DI,[pNumRows]                 ; tell the caller
              MOV       [WORD DI],NUMROWS             ;  ... what we've got
              MOV       DI,[pNumCols]
              MOV       [WORD DI],NUMCOLS
              MOV       DI,[pNumBytes]
              MOV       [WORD DI],TOTALCLOCKS         ; clocks = bytes!
              MOV       DI,[ppName]
              MOV       [WORD DI],OFFSET LCDName      ; set up string pointer

;--- clear the buffer

              MOV       [WORD PTR pLCDBuff],0         ; set up LCD pointer
              MOV       [WORD PTR pLCDBuff+2],GLCD_SEGMENT

              LES       DI,[pLCDBuff]
              MOV       AL,FILLVALUE
              MOV       CX,TOTALCLOCKS
              REP STOSB

;--- turn on the panel to show blank screen

              MOV       AX,GLCD_BLZERO      ; gate binary zero to data mux
              OR        AL,GLCD_ON          ; enable the output
              MOV       DX,GLCD_CTLS
              OUT       DX,AL               ; turn the display on
              MOV       [LCDCtlBits],AX     ; save for next time

;--- fill the row starting address table
;    this creates the appropriate RAM address for each row
;     ... the top row is wrapped from the bottom due to Frame Sync offset

              MOV       [RowStarts],(ROWMODULUS-1)*COLCLOCKS
              MOV       [RowStarts+2*ROWMODULUS],(ROWMODULUS-1)*COLCLOCKS

              MOV       DI,(OFFSET RowStarts) + 2
              MOV       CX,ROWMODULUS-1
              MOV       AX,0                          ; row 0 = second entry
              PUSH      DS                            ; set up STOSW segment
              POP       ES

@@1:          MOV       [DI+2*ROWMODULUS],AX          ; second half of table
              STOSW                                   ; first half
              ADD       AX,COLCLOCKS                  ; offset of next row
              LOOP      @@1

              RET

              ENDP      LCDInit

;-------------
; Create the LCD buffer address given row and column numbers
; Note that we pass parameters in registers rather than on the stack...
;  AX = column number, BX = row number
; Returns DI set to the byte offset within the buffer
;         CL contains shift needed to move the bit right into LSB (0..7)
; You must set up a seg reg and add the base offset to DI to get a pointer

              PROC      LCDMakeAddr

              ADD       BX,BX               ;  make word table index
              MOV       DI,[RowStarts+BX]

              DIV       [BYTE ColModulus]
              XOR       AH,03h              ; flip shift direction
              MOV       CL,AH               ; save remainder for shifting
              MOV       AH,0
              ADD       DI,AX               ; DI points to the byte

              CMP       BX,2*ROWMODULUS     ; rows >=200 use high nybble
              JB        @@1
              ADD       CL,4                ; ... so shift into it
@@1:

              RET

              ENDP      LCDMakeAddr

;-------------
; Plot a point on the LCD panel
; This always uses the buffer pointed to by pNewBuff

              PUBLIC    C LCDSetDot
              PROC      LCDSetDot
              ARG       Row:WORD,Col:WORD,Value:WORD
              USES      ES,DI

              MOV       AX,[Col]
              MOV       BX,[Row]
              CALL      LCDMakeAddr

              LES       AX,[pNewBuff]       ; get buffer base pointer
              ADD       DI,AX               ;  ... update byte pointer

              MOV       BX,0101h            ; set up bit mask
              SHL       BX,CL
              NOT       BL                  ;  BL = AND mask, BH = OR mask

              TEST      [Value],0001h       ; what do we want?
              JNZ       @@2                 ; nonzero means set the bit
              AND       [ES:DI],BL          ;  ... clear it
              JMP       SHORT @@3

@@2:
              OR        [ES:DI],BH          ;  ... set it

@@3:
              RET

              ENDP      LCDSetDot

;-------------
; Fetch a point from the LCD panel
; Always uses the buffer pointed to by pOldBuff
; We return 0 for any addresses outside the current panel
;  ... which greatly simplifies generating dots for The Game of Life!

              PUBLIC    C LCDGetDot
              PROC      LCDGetDot
              ARG       Row:WORD,Col:WORD
              USES      ES,DI

              XOR       AX,AX               ; set up zero return value

              MOV       BX,[Col]            ; verify column
              TEST      BH,80h              ; negative?
              JNZ       @@Done              ;  ... yes
              CMP       BX,NUMCOLS          ; too big?
              JAE       @@Done              ;  ... yes

              MOV       BX,[Row]            ; verify row
              TEST      BH,80h              ; negative?
              JNZ       @@Done              ;  ... yes
              CMP       BX,NUMROWS          ; too big?
              JAE       @@Done              ;  ... yes

              MOV       AX,[Col]            ; set up column number
              CALL      LCDMakeAddr

              LES       AX,[pOldBuff]       ; get buffer base pointer
              ADD       DI,AX               ;  ... update byte pointer

              MOV       AL,[ES:DI]          ;  ... fetch the byte
              SHR       AL,CL               ; skid our bit into LSB
              AND       AX,0001h            ;  ... isolate it
@@Done:
              RET

              ENDP      LCDGetDot

              END
