/* ANSI.C                                                                    */

#include <dos.h>
#include <ctype.h>
#include <mem.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "8086base.h"
#include "firmdev.h"
#include "utility.h"
#include "console.h"

#include "lcdplot.h"
#include "lcdchars.h"
#include "lcdansi.h"

/*---------------------------------------------------------------------------*/
/* Local prototypes                                                          */

int LCDAnsiGetNumber(void);
void LCDAnsiNull(int NewChar);
void LCDAnsiDoCmd(int NewChar);
void LCDAnsiSendString(char *pString);

/*---------------------------------------------------------------------------*/
/* Locally interesting constants                                             */


#define ANSI_MAXLEN     64             /* longest possible sequence!         */
#define ANSI_MAXNUMS    8              /* most numeric values                */

#define TABINTERVAL     4              /* for escape sequence tabbing        */

/*--- state machine control variable                                         */

#define ANSI_OFF     0                 /* no ANSI actions                    */
#define ANSI_WAIT    1                 /* waiting for ESC                    */
#define ANSI_MODE    2                 /* waiting for [                      */
#define ANSI_PARAM   3                 /* waiting for parameters             */
#define ANSI_NUM     4                 /* building numbers                   */

/*--- trace bits                                                             */

#define TRACE_ERRORS    0x0001         /* trace error conditions             */
#define TRACE_DECODE    0x0002         /* trace decoder paths                */
#define TRACE_CMDS      0x0004         /* trace command actions              */
#define TRACE_NUM       0x0008         /* trace numeric extraction           */
#define TRACE_NULL      0x0010         /* trace null command dump            */
#define TRACE_ECHO      0x0080         /* echo chars to serial port          */

/*---------------------------------------------------------------------------*/
/* Local globals...                                                          */

static int  ANSITrace;                 /* bitmap controlling trace outputs   */
static int  ANSIState;

static int  ANSICharCtr;
static int  ANSIParamCtr;

static char ANSIBuffer[ANSI_MAXLEN];
static int  ANSIParams[ANSI_MAXNUMS];

static int  ANSISavedRow;
static int  ANSISavedCol;

/*---------------------------------------------------------------------------*/
/* Initialize the ANSI sequence decoder                                      */

void LCDAnsiInit(void) {

   ANSIState = ANSI_WAIT;
   ANSICharCtr = ANSIParamCtr = 0;

   ANSISavedRow = ANSISavedCol = 1;

   LCDSetVScroll(0);                   /* do not allow screen scrolling!     */

}



/*---------------------------------------------------------------------------*/
/* Set tracing mode                                                          */

void LCDAnsiTrace(int Trace) {

   ANSITrace = Trace;

}



/*---------------------------------------------------------------------------*/
/* Accept a character and drive the state machine                            */

void LCDAnsiDriver(int NewChar) {

   if (0 != (ANSITrace & TRACE_DECODE)) {
      printf("--- ANSI char is %02x [%c], state %u, ctr %u\n",
         NewChar,isprint(NewChar) ? NewChar : '-',
         ANSIState,ANSICharCtr);
   }

   if (0 != (ANSITrace & TRACE_ECHO)) {
      putchar(NewChar);
   }

   if (ANSI_OFF == ANSIState) {        /* if driver off,                     */
      LCDSendChar(NewChar);            /* send directly to LCD panel         */
      return;                          /* and be done with it...             */
   }

   ANSIBuffer[ANSICharCtr] = NewChar & 0x7F;       /* strip & save           */
   ++ANSICharCtr;
   ANSICharCtr = min(ANSICharCtr,ANSI_MAXLEN-1);

/*--- decide what to do with char based on current state                     */

   switch (ANSIState) {
   case ANSI_WAIT :                    /* waiting for ESC char               */
      ANSICharCtr = ANSIParamCtr = 0;              /* force buffer restart   */
      if (ESC == NewChar) {
         if (0 != (ANSITrace & TRACE_DECODE)) {
            puts("   Start ANSI sequence");
         }
         memset(ANSIBuffer,0,sizeof(ANSIBuffer));
         ANSIState = ANSI_MODE;
      }
      else {
         LCDSendChar(NewChar);         /* just dump it                       */
      }
      break;
   case ANSI_MODE :                    /* expect bracket as next char        */
      if ('[' == NewChar) {
         ANSIState = ANSI_PARAM;
      }
      else {
         if (0 != (ANSITrace & TRACE_ERRORS)) {
            puts("    Missing [ char, sequence cancelled");
         }
         LCDAnsiNull(NewChar);                     /* ditch sequence         */
      }
      break;
   case ANSI_PARAM :
      if ('=' == NewChar) {
         break;                                    /* gobble this one...     */
      }
      else {
         if (isdigit(NewChar)) {
            ANSIState = ANSI_NUM;                  /* still a number         */
         }
         else {
            if (isalpha(NewChar)) {
               LCDAnsiGetNumber();                 /* get number if any      */
               LCDAnsiDoCmd(NewChar);              /* end of sequence        */
            }
            else {
               if (0 != (ANSITrace & TRACE_ERRORS)) {
                  printf("    Bad char %02x [%c], cancelled\n",
                         NewChar, isprint(NewChar) ? NewChar : '-');
               }
               LCDAnsiNull(NewChar);
            }
         }
      }
      break;
   case ANSI_NUM :
      if (';' == NewChar) {                        /* end of param digit     */
         LCDAnsiGetNumber();                       /* save prev param        */
      }
      else {
         if (isdigit(NewChar)) {
            ANSIState = ANSI_NUM;                  /* still a number         */
         }
         else {
            if (isalpha(NewChar)) {
               LCDAnsiGetNumber();                 /* get number if any      */
               LCDAnsiDoCmd(NewChar);              /* end of sequence        */
            }
            else {
               if (0 != (ANSITrace & TRACE_ERRORS)) {
                  printf("    Bad char %02x [%c] after param, cancelled\n",
                         NewChar, isprint(NewChar) ? NewChar : '-');
               }
               LCDAnsiNull(NewChar);
            }
         }
      }
      break;
   default :
      LCDAnsiInit();                               /* error, restart decoder */
   }


}

/*---------------------------------------------------------------------------*/
/* Extract a decimal number from the stored escape sequence                  */
/* Stored sequence included the char that triggered this, so we must back    */
/*  up over that character to begin the search                               */
/* Updates the array of parameters for easy access                           */

int LCDAnsiGetNumber(void) {
int TempNum;
char *pChar;

/* find start of numeric string                                              */

   if (0 != (ANSITrace & TRACE_NUM)) {
      printf("   Extracting param from [%s] @ %u: ",
         ANSIBuffer,ANSICharCtr-2);
   }

   for (pChar = ANSIBuffer + ANSICharCtr - 2;
        pChar >= ANSIBuffer;
        --pChar) {
      if (!isdigit(*pChar)) {
         ++pChar;
         if (0 != (ANSITrace & TRACE_NUM)) {
            printf("(Num string is [%s])  ",
               pChar);
         }
         break;
      }
   }

/* do the conversion                                                         */
/* if the number is missing, we default to 1 rather than zero                */

   if (isdigit(*pChar)) {
      TempNum = atoi(pChar);
   }
   else {
      TempNum = 1;
   }

/* update the array                                                          */

   ANSIParams[ANSIParamCtr] = TempNum;
   ++ANSIParamCtr;
   ANSIParamCtr = min(ANSIParamCtr,ANSI_MAXNUMS-1);


   if (0 != (ANSITrace & TRACE_NUM)) {
      printf(" %u\n",TempNum);
   }

   return TempNum;

}



/*---------------------------------------------------------------------------*/
/* Driver state routines                                                     */

/*--- mismatch handler... dump stored chars & reset to beginning             */

void LCDAnsiNull(int NewChar) {

   NewChar = NewChar;                  /* do not need this, already stored   */

   if (ANSICharCtr) {
      if (0 != (ANSITrace & TRACE_NULL)) {
         printf("   Dumping sequence [%s]\n",ANSIBuffer);
      }
      LCDSendChar(ESC);                /* not stored in buffer               */
      LCDSendString(ANSIBuffer);       /* all chars including current one    */
   }

   ANSICharCtr = ANSIParamCtr = 0;
   ANSIState = ANSI_WAIT;

}



/*--- hit end of sequence, so do the appropriate command action routine      */

void LCDAnsiDoCmd(int NewChar) {

BYTE Counter;

   if (0 != (ANSITrace & TRACE_CMDS)) {
      printf("   Start decode:  ");
   }

   ANSIState = ANSI_WAIT;                          /* by default...          */

   switch (NewChar) {
   case 'A' :                          /* cursor up                          */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Cursor up %u ",
            ANSIParams[0]);
      }
      LCDSetCharCursor(LCDGetCharRow()-ANSIParams[0],LCDGetCharCol());
      break;
   case 'B' :                          /* cursor down                        */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Cursor down %u ",
            ANSIParams[0]);
      }
      LCDSetCharCursor(LCDGetCharRow()+ANSIParams[0],LCDGetCharCol());
      break;
   case 'C' :                          /* cursor forward                     */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Cursor forward %u ",
            ANSIParams[0]);
      }
      LCDSetCharCursor(LCDGetCharRow(),LCDGetCharCol()+ANSIParams[0]);
      break;
   case 'D' :                          /* cursor back                        */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Cursor back %u ",
            ANSIParams[0]);
      }
      LCDSetCharCursor(LCDGetCharRow(),LCDGetCharCol()-ANSIParams[0]);
      break;
   case 'H' :                          /* set cursor position                */
   case 'f' :                          /* set cursor position (synonym)      */
   case 'F' :                          /* set cursor position (synonym)      */
      for (;ANSIParamCtr < 2; ++ANSIParamCtr) {
         ANSIParams[ANSIParamCtr] = 1;             /* force defaults         */
      }
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Cursor to ANSI row %u, col %u ",
            ANSIParams[0],ANSIParams[1]);
      }
      LCDSetCharCursor(ANSIParams[0]-1,ANSIParams[1]-1);    /* fix origin!   */
      break;
   case 'J' :                          /* erase display & home cursor        */
      if (2 == ANSIParams[0]) {
         if (0 != (ANSITrace & TRACE_CMDS)) {
            printf("   Clear and home cursor");
         }
         LCDClear();
      }
      break;
   case 'K' :                          /* erase to end of line               */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Clear to end of line");
      }
      Counter = NumCharCols - LCDGetCharCol();
      while (Counter) {
         LCDSendChar(DEL);
         --Counter;
      }
      break;
   case 'R' :                          /* cursor location report response    */
      break;                           /*  ... ignored                       */
   case 'h' :                          /* set screen mode, wrap at line end  */
      if ((ANSIParamCtr == 1) && (ANSIParams[0] == 7)) {
         if (0 != (ANSITrace & TRACE_CMDS)) {
            printf("   Set line wrap mode ON");
         }
         LCDSetHWrap(1);
      }
      break;                           /*  ... ignored                       */
   case 'l' :                          /* set screen mode, no wrap at end    */
      if ((ANSIParamCtr == 1) && (ANSIParams[0] == 7)) {
         if (0 != (ANSITrace & TRACE_CMDS)) {
            printf("   Set line wrap mode OFF");
         }
         LCDSetHWrap(0);
      }
      break;                           /*  ... ignored                       */
   case 'm' :                          /* set graphics rendition             */
      if (ANSIParamCtr == 1) {
         switch (ANSIParams[0]) {
         case 0 :
            if (0 != (ANSITrace & TRACE_CMDS)) {
               printf("   Set blinking OFF");
            }
            LCDSetBlink(0);
            break;
         case 7 :
            if (0 != (ANSITrace & TRACE_CMDS)) {
               printf("   Set blinking ON");
            }
            LCDSetBlink(1);
            break;
         default :
            if (0 != (ANSITrace & TRACE_ERRORS)) {
               printf("*** attr not supported: %u",ANSIParams[0]);
            }
         }
      }
      break;
   case 'n' :                          /* query cursor location              */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Query cursor location (row %u, col %u)",
                  LCDGetCharRow()+1,LCDGetCharCol()+1);
      }
/*    we do not return this to the caller in stdin() as required...          */
/*    printf("\x1b[%u;%uR\n",LCDGetCharRow()+1,LCDGetCharCol()+1);           */
      break;
   case 'p' :                          /* reassign key values                */
      break;
   case 's' :                          /* save cursor location               */
      ANSISavedRow = LCDGetCharRow()+1;
      ANSISavedCol = LCDGetCharCol()+1;
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Save cursor (ANSI row %u, col %u)",
                  ANSISavedRow,ANSISavedCol);
      }
      break;
   case 'u' :                          /* restore cursor location            */
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Restore cursor (ANSI row %u, col %u)",
                  ANSISavedRow,ANSISavedCol);
      }
      LCDSetCharCursor(ANSISavedRow-1,ANSISavedCol-1);
      break;
   default :
      if (0 != (ANSITrace & TRACE_CMDS)) {
         printf("   Invalid command char [%c] %02x hex",
                  iscntrl(NewChar)?NewChar:'-',NewChar);
      }
      LCDAnsiNull(NewChar);                        /* dump all others...     */
   }

   if (0 != (ANSITrace & TRACE_CMDS)) {
      puts("");
   }

}


/*---------------------------------------------------------------------------*/
/* Send a string to the ANSI decoder                                         */
/* We decode a few C-style formatting characters...                          */
/*  to simplify life for non-C applications                                  */

void LCDAnsiSendString(char *pString) {

char RepString[10];
char *pRep;

   while (*pString) {
      if ('\\' == *pString) {                      /* escape sequence?       */
         switch (tolower(*(pString+1))) {
         case 'b' :                                /* bell sound             */
            RepString[0] = 0;                      /*  ... not to LCD        */
            break;
         case 'c' :                                /* control chars          */
            RepString[0] = *(pString+2) & 0x1F;
            RepString[1] = 0;
            pString += 1;                          /* account for char       */
            break;
         case 'e' :                                /* escape char itself     */
            strcpy(RepString,"\x1B");
            break;
         case 'f' :                                /* formfeed               */
            strcpy(RepString,"\x0C");
            break;
         case 'n' :                                /* newline (LF and CR)    */
            strcpy(RepString,"\x0A");
            break;
         case 'r' :                                /* bare CR                */
            strcpy(RepString,"\x0D");
            break;
         case 't' :                                /* tab                    */
            strcpy(RepString,"\x09");
            break;
         case '\\' :
            RepString[0] = '\\';                   /* send single slash      */
            RepString[1] = 0;
            break;
         default :
            RepString[0] = *(pString+1);           /* send next char anyway  */
            RepString[1] = 0;
         }
         pString += 2;
         pRep = RepString;
         while (*pRep) {
            LCDAnsiDriver(*pRep);
            ++pRep;
         }
      }
      else {
         LCDAnsiDriver(*pString);                  /* just copy the char     */
         ++pString;
      }


   }

}

