/* Character and string output for Graphic LCD Interface                     */

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>

#include "8086base.h"
#include "firmdev.h"
#include "utility.h"

#include "LCDplot.h"
#include "LCDChars.h"

/*---------------------------------------------------------------------------*/
/* Globals that everybody needs to know                                      */

#define TABINTERVAL     4              /* characters per tab stop            */

unsigned int NumFontChars;             /* number of chars in font table      */
int FontHeight;                        /* number of rows in each char        */
int FontWidth;                         /*  ... and columns                   */

int NumCharRows;                       /* panel size in characters           */
int NumCharCols;

/*---------------------------------------------------------------------------*/
/* Variables                                                                 */

static int BlinkMode;                  /* 1 = now blinking                   */

static BYTE far *pFont;                /* points to first char in font table */

static int VScrollAllowed;             /* 1 = vert scroll at lower right     */
static int HWrapAllowed;               /* 1 = wrap at end of line            */
static int CurDotRow;                  /* cursor location in dot units       */
static int CurDotCol;

/*---------------------------------------------------------------------------*/
/* Initialize the Graphics LCD Panel character generator                     */
/* Uses VGA font if available, otherwise BIOS CGA font                       */
/* If Int 1F vector is null, we must use the CGA font for chars 00..7F only  */
/*   otherwise, invoke Int 10 AH=11 AL=30 to get font information            */
/*   this returns the =current= font height, not the height pointed to!      */

void LCDInitFont(void) {

struct REGPACK rp;
void far *Int1FVector;
int VidMode;

/*--- figure out what character table is available                           */

   puts("Initializing character generator...");

   Int1FVector = (void far *)*(long far *)MK_FP(0x0000,4*0x1F);
   printf("  Int 1F vector is %Fp\n",Int1FVector);
   if ((void far *)NULL == Int1FVector) {          /* have chars 80..FF?     */
      puts("  Using ROM BIOS CGA font");
      pFont = (BYTE far *)MK_FP(0xF000,0xFA6E);    /*  nope, use BIOS ROM    */
      FontHeight = 8;                              /*  which is 8x8 only     */
      FontWidth = 8;
      NumFontChars = 128;                          /*  basic ASCII only!     */
   }
   else {
      puts("  Using video card font");
      VidMode = *(BYTE far *)MK_FP(0x0040,0x0049);
      printf("  ... resetting video mode: %u\n",VidMode);

      rp.r_ax = (0x00 << 8) | VidMode;             /* Reset video mode       */
      intr(0x10,&rp);

      rp.r_ax = 0x1130;                            /* Get Font Pointer info  */
      if (NumDotRows > 200) {
         rp.r_bx = 0x0600;                         /*  BH = 06 for 8x16 font */
         FontHeight = 16;
      }
      else {
         rp.r_bx = 0x0300;                         /*  BH = 03 for 8x8 font  */
         FontHeight = 8;
      }
      printf("  ... fetching %u line font pointer\n",FontHeight);
      intr(0x10,&rp);

#if 0
      printf("ax %04X bx %04X cx %04X dx %04X ds %04X es %04X bp %04X\n",
            rp.r_ax,rp.r_bx,rp.r_cx,rp.r_dx,rp.r_ds,rp.r_es,rp.r_bp);
#endif

      pFont = (BYTE far *)MK_FP(rp.r_es,rp.r_bp);
      FontWidth = 8;
      NumFontChars = 256;
   }

   printf("  pFont=%Fp, FontHeight=%u, NumFontChars=%u\n",
          pFont,FontHeight,NumFontChars);

/*--- set up screen control variables                                        */

   NumCharRows = NumDotRows / FontHeight;
   NumCharCols = NumDotCols / FontWidth;
   VScrollAllowed = 1;

   printf("  Panel has %u character rows and %u character cols\n",
          NumCharRows,NumCharCols);

   return;

}


/*---------------------------------------------------------------------------*/
/* Write a single glyph to the LCD buffer                                    */
/* Uses current font table and blinking mode                                 */
/* Coordinates are in dot rows and columns                                   */

void LCDWriteGlyph(int CharCode,int DotRow,int DotCol) {

int ScanLine;

   if ((CharCode < 0) || (CharCode >= NumFontChars)) {
      CharCode = '?';                  /* an obvious ASCII filler character  */
   }

   CharCode *= FontHeight;                         /* point to start of char */

   for (ScanLine=0; ScanLine < FontHeight; ++ScanLine) {
      LCDWriteByte(*(pFont+CharCode+ScanLine),
                   DotRow+ScanLine,DotCol,BlinkMode);
   }

}


/*---------------------------------------------------------------------------*/
/* Set cursor position to specified dot row and column                       */

void LCDSetDotCursor(int DotRow,int DotCol) {

   DotRow = max(DotRow,0);
   DotRow = min(DotRow,NumDotRows-1);
   CurDotRow = DotRow;

   DotCol = max(DotCol,0);
   DotCol = min(DotCol,NumDotCols-1);
   CurDotCol = DotCol;


}


/*---------------------------------------------------------------------------*/
/* Set cursor position to specified character row and column                 */

void LCDSetCharCursor(int Row,int Col) {

int DotRow,DotCol;

   DotRow = max(Row,0) * FontHeight;
   DotRow = min(DotRow,NumDotRows-1);
   CurDotRow = DotRow;

   DotCol = max(Col,0) * FontWidth;
   DotCol = min(DotCol,NumDotCols-1);
   CurDotCol = DotCol;


}

/*---------------------------------------------------------------------------*/
/* Return current cursor dot location                                        */

void LCDGetDotCursor(int *pDotRow,int *pDotCol) {

   *pDotRow = CurDotRow;
   *pDotCol = CurDotCol;

}


/*---------------------------------------------------------------------------*/
/* Return current cursor character row                                       */

int  LCDGetCharRow(void) {

   return (int)CurDotRow / FontHeight;

}

/*---------------------------------------------------------------------------*/
/* Return current cursor character column                                    */

int  LCDGetCharCol(void) {

   return CurDotCol / FontWidth;

}


/*---------------------------------------------------------------------------*/
/* Set cursor to 0,0 but do not clear the panel                              */

void LCDHome(void) {

   LCDSetDotCursor(0,0);

}


/*---------------------------------------------------------------------------*/
/* Set cursor to 0,0 and clear the panel                                     */
/* This does not do a complete reset to avoid showing startup glitches       */

void LCDClear(void) {

   LCDEmpty();                         /* fill buffer with "all off" code    */
   LCDHome();

}



/*---------------------------------------------------------------------------*/
/* Set blinking mode ON or OFF                                               */

void LCDSetBlink(int Blink) {

   BlinkMode = BlinkAllowed ? Blink : 0;

}


/*---------------------------------------------------------------------------*/
/* Set scrolling mode ON or OFF                                              */

void LCDSetVScroll(int Scroll) {

   VScrollAllowed = Scroll;

}

/*---------------------------------------------------------------------------*/
/* Set wrap mode ON or OFF                                                   */

void LCDSetHWrap(int Wrap) {

   HWrapAllowed = Wrap;

}


/*---------------------------------------------------------------------------*/
/* Send printable chars to the LCD                                           */
/* Handles standard control chars in a reasonable manner                     */
/* Coordinates are in terms of dot rows and columns                          */

void LCDSendChar(int CharCode) {

   switch (CharCode) {
   case LF :
      if (CurDotRow >= (NumDotRows - FontHeight)) {         /* on last line? */
         if (VScrollAllowed) {
            LCDScrollUp(FontHeight);               /* yes, scroll up a line  */
            CurDotRow = NumDotRows - FontHeight;   /* force to last line     */
         }
         else {
            CurDotRow = 0;                         /* no scroll, wrap to top */
         }
      }
      else {
         CurDotRow += FontHeight;                 /* no, so step down        */
      }
      break;

   case CR :
      CurDotCol = 0;                               /* to leftmost column     */
      break;

   case FORMFEED :
      LCDClear();
      break;

   case BS :
      if (CurDotCol >= FontWidth) {                /* if not at edge         */
         CurDotCol -= FontWidth;                   /*   back up one col      */
      }
      break;

   case TAB :                                      /* tab to next stop       */
      do {
         LCDSendChar(' ');
      } while (((CurDotCol/FontWidth)+1) % TABINTERVAL);
      break;

   case DEL :
      LCDScrollLeft(CurDotRow,CurDotCol,FontHeight,FontWidth);
      break;

   default :
      LCDWriteGlyph(CharCode,CurDotRow,CurDotCol); /* display it             */
      CurDotCol += FontWidth;                      /* tick column counter    */
      if (CurDotCol > (NumDotCols - FontWidth)) {  /* past end of line?      */
         CurDotCol = 0;                            /* reset to left edge     */
         if (!HWrapAllowed) {
            LCDSendChar(LF);                       /* and step to next line  */
         }
      }
   }

}



/*---------------------------------------------------------------------------*/
/* Send string to LCD starting at current cursor position                    */

void LCDSendString(char *pString) {

   while (0 != *pString) {
      LCDSendChar(*pString);
      ++pString;
   }

}

