; Graphic LCD Interface routines for 640x400 8-bit 480 ns panel
; This has 200 rows of 160 clocks

              IDEAL
              P386N

              INCLUDE   "firmdev.inc"

              MODEL     SMALL,CPP

;-------------
; Define LCD parameters
; We must use = rather than EQU to get real numeric results!

LCDNAME       EQU       "Matsushita EDM LG64AA44D"
NUMROWS       =         400
NUMCOLS       =         640
COLCLOCKS     =         160
COLMODULUS    =         4
TOTALCLOCKS   =         32000
ROWMODULUS    =         NUMROWS/2
FILLVALUE     =         000h                          ; fill both nybbles!
BLINKALLOWED  =         0

              EXTRN     C Load8254:PROC

;-------------
; Set up data values needed by the plotting routines

              DATASEG

              PUBLIC    C NumDotBytes
              PUBLIC    C NumDotRows
              PUBLIC    C NumDotCols
              PUBLIC    C BlinkAllowed
              PUBLIC    C pLCDName

NumDotBytes   DW        TOTALCLOCKS
NumDotRows    DW        NUMROWS
NumDotCols    DW        NUMCOLS
BlinkAllowed  DW        BLINKALLOWED
pLCDName      DD        LCDName             ; pick up the address

pLCDBuff      DD        ?                   ; pointer to LCD buffer

LCDName       DB        LCDNAME             ; fill in the name
              DB        0                   ;  ... and a terminator

ColModulus    DW        COLMODULUS          ; convert dot col to byte offset

RowStarts     DW        NUMROWS DUP (?)     ; row starting addresses

LCDCtlBits    DW        0                   ; copy of GLCD_CTLS output port

              CODESEG

              EXTRN     _Load8254

;-------------
; Enable & disable the LCD panel

              PUBLIC    C LCDEnable
              PROC      LCDEnable
              ARG       Enable : WORD

              MOV       AX,[LCDCtlBits]     ; pick up current controls

              TEST      [Enable],0001h      ; 1 = turn it on...
              JNZ       SHORT @@On

              AND       AX,NOT GLCD_ENDISP  ; lower the enable line
              JMP       SHORT @@Set
@@On:
              OR        AX,GLCD_ENDISP      ; raise the enable line

@@Set:
              MOV       DX,GLCD_CTLS        ; tell the port
              OUT       DX,AX

              MOV       [LCDCtlBits],AX     ; and save for next time

              RET

              ENDP      LCDEnable


;-------------
; Initialize the Graphic LCD Interface hardware
; Also fills the buffer and returns some useful numbers
; This keeps the initialization code in here where it belongs...

              PUBLIC    C LCDInit
              PROC      LCDInit
              USES      ES,DI

;--- disable the panel and set up the sync counters

              MOV       AL,GLCD_OFF         ; shut the interface down
              MOV       DX,GLCD_CTLS
              OUT       DX,AL

              CALL      Load8254 C,1,04h,COLCLOCKS    ; use Mode 2
              CALL      Load8254 C,2,04h,TOTALCLOCKS  ;  ... for both values

;--- clear the buffer

              MOV       [WORD PTR pLCDBuff],0         ; set up LCD pointer
              MOV       [WORD PTR pLCDBuff+2],GLCD_SEGMENT

              CALL      LCDEmpty C

;--- turn on the panel to show blank screen

              MOV       AX,GLCD_BLZERO      ; gate binary zero to data mux
              OR        AL,GLCD_ON          ; enable the output
              MOV       DX,GLCD_CTLS
              OUT       DX,AL               ; turn the display on
              MOV       [LCDCtlBits],AX     ; save for next time

;--- fill the row starting address table
;    this creates the appropriate RAM address for each row
;     ... the top row is wrapped from the bottom due to Frame Sync offset

              MOV       [RowStarts],(ROWMODULUS-1)*COLCLOCKS
              MOV       [RowStarts+2*ROWMODULUS],(ROWMODULUS-1)*COLCLOCKS

              MOV       DI,(OFFSET RowStarts) + 2
              MOV       CX,ROWMODULUS-1
              MOV       AX,0                          ; row 0 = second entry
              PUSH      DS                            ; set up STOSW segment
              POP       ES

@@1:          MOV       [DI+2*ROWMODULUS],AX          ; second half of table
              STOSW                                   ; first half
              ADD       AX,COLCLOCKS                  ; offset of next row
              LOOP      @@1

              RET

              ENDP      LCDInit

;-------------
; Fill the LCD buffer with the background value
; This does =not= use the current blinking mode, it just fills the buffer

              PUBLIC    C LCDEmpty
              PROC      LCDEmpty
              USES      ES,DI

              LES       DI,[pLCDBuff]       ; set up buffer pointer
              MOV       AL,FILLVALUE        ; fetch the default fill value
              MOV       CX,TOTALCLOCKS      ; fill only the affected area
              CLD
              REP STOSB

              RET

              ENDP      LCDEmpty


;-------------
; Create the LCD buffer address given row and column numbers
; Note that we pass parameters in registers rather than on the stack...
;  AX = column number, BX = row number
; Returns DI set to the byte offset within the buffer
;         CL contains shift needed to move the nybble right into the LSB
; You must set up a seg reg and add the base offset to DI to get a pointer
; CL is either 0 or 4 depending on which half of the panel the char is in

              PROC      LCDMakeCharAddr

              ADD       BX,BX               ;  make word table index
              MOV       DI,[RowStarts+BX]

              DIV       [BYTE ColModulus]
              MOV       CL,0                ; assume top of panel = no shift
              MOV       AH,0
              ADD       DI,AX               ; DI points to the byte

              CMP       BX,2*ROWMODULUS     ; rows >=200 use high nybble
              JB        SHORT @@1
              ADD       CL,4                ; low half, need shift
@@1:

              RET

              ENDP      LCDMakeCharAddr

;-------------
; Write a byte into the LCD panel buffer
; Forces byte alignment by truncating column number
;  ... uses multiple of 8 to ensure whole character fits on one line
; The Blinking parameter is accepted for compatibility with other panels,
;   but it's ignored here because we can't blink this one...

              PUBLIC    C LCDWriteByte
              PROC      LCDWriteByte
              ARG       CharCode:WORD,DotRow:WORD,DotCol:WORD,Blinking:WORD
              USES      ES,DI

              MOV       AX,[DotCol]         ; set up column number
              AND       AX,NOT 0007h        ;  ... force byte alignment
              MOV       BX,[DotRow]         ; set up row number
              CALL      LCDMakeCharAddr

              LES       AX,[pLCDBuff]       ; get buffer base pointer
              ADD       DI,AX               ;  ... update byte pointer

;--- insert the left nybble

              MOV       AX,[CharCode]       ; fetch new byte value
              MOV       DL,AL               ; save high bits for later
              SHR       AL,4                ; set up for alignment
              SHL       AL,CL               ; align to target nybble
              MOV       BH,0F0h             ; set up bit mask
              ROL       BH,CL               ;  ... align to strip old bits

              MOV       AH,[ES:DI]          ; fetch target bits
              AND       AH,BH               ; strip old bits
              OR        AL,AH               ; tuck in new bits
              MOV       [ES:DI],AL          ; put 'em back in the buffer

;--- insert the right nybble
;    this uses the same nybble mask because it's in the same screen half

              INC       DI                  ; step to next buffer byte
              AND       DL,00Fh             ; extract low nybble
              SHL       DL,CL               ; align to target nybble

              AND       BH,[ES:DI]          ; fetch target bits, strip old
              OR        DL,BH               ; tuck in new bits
              MOV       [ES:DI],DL          ; put 'em back in the buffer

              RET

              ENDP      LCDWriteByte

;-------------
; Scroll the display upward by specified number of dot rows
; Injects that number of blank dot rows on the bottom of the panel

              PUBLIC    C LCDScrollUp
              PROC      LCDScrollUp
              ARG       DotRows:WORD
              USES      ES,FS,DI,SI
              LOCAL     TopRow:WORD,BotRow:WORD

              XOR       AX,AX               ; target is top row, always 0
              MOV       [TopRow],AX

              MOV       CX,[DotRows]        ; source row = number of rows
              MOV       [BotRow],CX

              CMP       CX,0                ; bail out if count is zero
              JE        SMALL @@Done

;--- loop for each scan line to be moved up
;    terminates when last scan line is just beyond bottom of panel

@@1:
              CMP       [BotRow],NUMROWS    ; source - bottom row
              JAE       SHORT @@NoMove      ;  if source is beyond bottom, done

              XOR       AX,AX               ; set up source pointer in FS:SI
              MOV       BX,[BotRow]
              CALL      LCDMakeCharAddr
              LFS       SI,[pLCDBuff]
              ADD       SI,DI
              PUSH      CX                  ; save source shift count

              XOR       AX,AX               ; set up target pointer in ES:DI
              MOV       BX,[TopRow]
              CALL      LCDMakeCharAddr
              LES       AX,[pLCDBuff]
              ADD       DI,AX

              POP       AX                  ; recover source shift count
              MOV       CH,AL               ; set it up in CH for loop

              PUSH      COLCLOCKS           ; push the counter value

@@2:
              MOV       AH,[ES:DI]          ; fetch target nybble
              MOV       AL,0F0h             ; set up mask for target byte
              ROL       AL,CL               ; align it
              AND       AH,AL               ;  ... remove target nybble

              MOV       AL,[FS:SI]          ; fetch source byte
              XCHG      CH,CL               ; get source shift count
              SHR       AL,CL               ; align bits in low nybble
              AND       AL,00Fh             ; strip off unused bits
              XCHG      CH,CL               ; get target shift count
              SHL       AL,CL               ; align bits for target byte
              OR        AL,AH               ;  ... tuck in the source nybble
              MOV       [ES:DI],AL          ;  ... and save the two

              INC       SI                  ; tick pointers
              INC       DI

              POP       AX                  ; recover the loop counter
              DEC       AX
              PUSH      AX
              JNZ       @@2

              POP       AX                  ; discard the counter

              INC       [TopRow]            ; step to next row
              INC       [BotRow]

              JMP       SHORT @@1           ; continue until done at top

@@NoMove:

;--- now fill in the vacated rows starting with TopRow
;    ends when TopRow falls off the bottom of the panel

@@3:
              CMP       [TopRow],NUMROWS    ;  current - total rows
              JAE       SHORT @@Done        ;  if current >= total, finished

              XOR       AX,AX               ; get target pointer in ES:DI
              MOV       BX,[TopRow]
              CALL      LCDMakeCharAddr
              LES       AX,[pLCDBuff]
              ADD       DI,AX

              MOV       BH,FILLVALUE AND 00Fh         ; set up fill nybble
              SHL       BH,CL               ; align to target location

              MOV       BL,0F0h             ; set up mask to strip target
              ROL       BL,CL               ; align to target location

              MOV       CX,COLCLOCKS        ; set up row counter

@@4:
              AND       [ES:DI],BL          ; strip target bits
              OR        [ES:DI],BH          ; tuck in fill bits

              INC       DI                  ; step to next buffer byte

              LOOP      @@4                 ; repeat for this row

              INC       [TopRow]            ; step row number
              JMP       SHORT @@3           ; do it again until done at top

@@Done:
              RET

              ENDP      LCDScrollUp

;-------------
; Move a block leftward by specified number of dot columns
; Injects that number of blank dots on the right edge of the block
; This routine handles only blocks that are multiples of four bits
;  ... to avoid partial byte shuffling that we don't need for characters

              PUBLIC    C LCDScrollLeft
              PROC      LCDScrollLeft
              ARG       DotRow:WORD,DotCol:WORD,Height:WORD,Width:WORD
              USES      ES,FS,DI,SI
              LOCAL     ByteWidth:WORD,FillWidth:WORD

;--- convert "move" and "fill" widths from dots to bytes

              MOV       AX,[DotCol]
              ADD       AX,[Width]          ; source dot moved to DotCol
              NEG       AX
              ADD       AX,NUMCOLS          ; last+1 - first dot = # to move
              DIV       [BYTE ColModulus]
              ADD       AH,0FFh             ; carry set if fraction of byte
              ADC       AL,0                ;  ... to round upward
              XOR       AH,AH               ; always less than 256 bytes!
              MOV       [ByteWidth],AX

              MOV       AX,[Width]
              DIV       [BYTE ColModulus]
              XOR       AH,AH               ; discard remainder to truncate
              MOV       [FillWidth],AX

;--- loop for each scan line in the overall Height region
;    this does not handle wraps off the end of the line very well at all...

@@1:
              CMP       [Height],0          ; check for doneness
              JLE       SHORT @@Done        ;  if <= 0 we are through

              MOV       AX,[DotCol]         ; set up source pointer in FS:SI
              ADD       AX,[Width]
              MOV       BX,[DotRow]
              CALL      LCDMakeCharAddr
              LFS       SI,[pLCDBuff]
              ADD       SI,DI
              PUSH      CX                  ; save source shift count

              MOV       AX,[DotCol]         ; set up target pointer in ES:DI
              MOV       BX,[DotRow]
              CALL      LCDMakeCharAddr
              LES       AX,[pLCDBuff]
              ADD       DI,AX

              POP       AX                  ; recover source shift count
              MOV       CH,CL               ;  ... set up for the loop

              PUSH      [ByteWidth]         ; set up move counter

@@2:
              MOV       AH,[ES:DI]          ; fetch target nybble
              MOV       AL,0F0h             ; set up mask for target byte
              ROL       AL,CL               ; align it
              AND       AH,AL               ;  ... remove target nybble

              MOV       AL,[FS:SI]          ; fetch source byte
              XCHG      CH,CL               ; get source shift count
              SHR       AL,CL               ; align bits in low nybble
              AND       AL,00Fh             ; strip off unused bits
              XCHG      CH,CL               ; get target shift count
              SHL       AL,CL               ; align bits for target byte
              OR        AL,AH               ;  ... tuck in the source nybble
              MOV       [ES:DI],AL          ;  ... and store them away

              INC       SI                  ; tick pointers
              INC       DI

              POP       AX                  ; recover the loop counter
              DEC       AX
              PUSH      AX
              JNZ       @@2

              POP       AX                  ; discard the counter

              MOV       BH,FILLVALUE AND 00Fh         ; set up fill nybble
              SHL       BH,CL               ; align to target location

              MOV       BL,0F0h             ; set up mask to strip target
              ROL       BL,CL               ; align to target location

              MOV       CX,[FillWidth]

@@3:
              AND       [ES:DI],BL          ; strip target bits
              OR        [ES:DI],BH          ; tuck in fill bits

              INC       DI                  ; step to next buffer byte

              LOOP      @@3                 ; repeat for this row

              INC       [DotRow]            ; step to next row
              DEC       [Height]            ; and tick row counter

              JMP       SHORT @@1           ; continue until done at top

@@Done:
              RET

              ENDP      LCDScrollLeft

              END
